import unittest

from gradgen import (
    CodeGenerationBuilder,
    Function,
    RustBackendConfig,
    SX,
    SXVector,
    SingleShootingBundle,
    SingleShootingProblem,
)

from gradgen.single_shooting import (
    _slice_packed_sequence,
    _extract_single_output,
    _extract_scalar_output,
    _flatten_rollout_states,
    _validate_single_shooting_bundle,
    _single_shooting_primal_name,
    _single_shooting_joint_name,
    _single_shooting_bundle_output_names,
    SingleShootingBundle,
    SingleShootingProblem,
)

def _build_minimal_scalar_problem():
    # state is 2-D, control is scalar, parameter is 3-D
    x = SXVector.sym("x", 2)
    u = SX.sym("u")
    p = SXVector.sym("p", 3)

    dynamics = Function(
        "d",
        [x, u, p],
        [SXVector((0.5 * x[0] + p[0] * u, 0.5 * x[1] + p[1] * u))],
    )
    stage_cost = Function(
        "s",
        [x, u, p],
        [x[0] * x[0] + x[1] * x[1] + u * u + p[2] * u],
    )
    terminal_cost = Function(
        "t",
        [x, p],
        [x[0] * x[0] + x[1] * x[1] + p[2] * x[1]],
    )
    return SingleShootingProblem(
        "sp",
        horizon=2,
        dynamics=dynamics,
        stage_cost=stage_cost,
        terminal_cost=terminal_cost,
    )

def _build_reference_problem(*, horizon: int = 3) -> SingleShootingProblem:
    x = SXVector.sym("x", 2)
    u = SXVector.sym("u", 1)
    p = SXVector.sym("p", 2)

    dynamics = Function(
        "dynamics",
        [x, u, p],
        [SXVector((x[0] + p[0] * x[1] + u[0], x[1] + p[1] * u[0] - 0.5 * x[0]))],
        input_names=["x", "u", "p"],
        output_names=["x_next"],
    )
    stage_cost = Function(
        "stage_cost",
        [x, u, p],
        [x[0] * x[0] + 2.0 * x[1] * x[1] + 0.3 * u[0] * u[0] + p[0] * u[0]],
        input_names=["x", "u", "p"],
        output_names=["ell"],
    )
    terminal_cost = Function(
        "terminal_cost",
        [x, p],
        [3.0 * x[0] * x[0] + 0.5 * x[1] * x[1] + p[1] * x[0]],
        input_names=["x", "p"],
        output_names=["vf"],
    )
    return SingleShootingProblem(
        name="mpc_cost",
        horizon=horizon,
        dynamics=dynamics,
        stage_cost=stage_cost,
        terminal_cost=terminal_cost,
        initial_state_name="x0",
        control_sequence_name="U",
        parameter_name="p",
    )


def _manual_rollout(x0: list[float], U: list[float], p: list[float], horizon: int) -> tuple[float, list[float]]:
    current = [float(x0[0]), float(x0[1])]
    packed_states = [current[0], current[1]]
    total_cost = 0.0

    for stage_index in range(horizon):
        u_t = float(U[stage_index])
        total_cost += current[0] * current[0] + 2.0 * current[1] * current[1] + 0.3 * u_t * u_t + p[0] * u_t
        next_state = [
            current[0] + p[0] * current[1] + u_t,
            current[1] + p[1] * u_t - 0.5 * current[0],
        ]
        current = next_state
        packed_states.extend(current)

    total_cost += 3.0 * current[0] * current[0] + 0.5 * current[1] * current[1] + p[1] * current[0]
    return total_cost, packed_states


def _finite_difference_gradient(
    x0: list[float],
    U: list[float],
    p: list[float],
    horizon: int,
    *,
    epsilon: float = 1e-6,
) -> list[float]:
    gradient: list[float] = []
    for control_index in range(len(U)):
        forward = list(U)
        backward = list(U)
        forward[control_index] += epsilon
        backward[control_index] -= epsilon
        forward_cost, _ = _manual_rollout(x0, forward, p, horizon)
        backward_cost, _ = _manual_rollout(x0, backward, p, horizon)
        gradient.append((forward_cost - backward_cost) / (2.0 * epsilon))
    return gradient


def _finite_difference_hvp(
    x0: list[float],
    U: list[float],
    p: list[float],
    v_U: list[float],
    horizon: int,
    *,
    epsilon: float = 1e-4,
) -> list[float]:
    forward_controls = [
        control + epsilon * direction for control, direction in zip(U, v_U)
    ]
    backward_controls = [
        control - epsilon * direction for control, direction in zip(U, v_U)
    ]
    forward_gradient = _finite_difference_gradient(x0, forward_controls, p, horizon, epsilon=epsilon)
    backward_gradient = _finite_difference_gradient(x0, backward_controls, p, horizon, epsilon=epsilon)
    return [
        (forward_value - backward_value) / (2.0 * epsilon)
        for forward_value, backward_value in zip(forward_gradient, backward_gradient)
    ]


def _build_scalar_reference_problem(*, horizon: int = 2) -> SingleShootingProblem:
    x = SX.sym("x")
    u = SX.sym("u")
    p = SX.sym("p")

    dynamics = Function(
        "scalar_dynamics",
        [x, u, p],
        [0.5 * x + p * u],
        input_names=["x", "u", "p"],
        output_names=["x_next"],
    )
    stage_cost = Function(
        "scalar_stage_cost",
        [x, u, p],
        [x * x + 2.0 * u * u + p * x * u],
        input_names=["x", "u", "p"],
        output_names=["ell"],
    )
    terminal_cost = Function(
        "scalar_terminal_cost",
        [x, p],
        [3.0 * x * x + p * x],
        input_names=["x", "p"],
        output_names=["vf"],
    )
    return SingleShootingProblem(
        name="scalar_cost",
        horizon=horizon,
        dynamics=dynamics,
        stage_cost=stage_cost,
        terminal_cost=terminal_cost,
        initial_state_name="x0",
        control_sequence_name="U",
        parameter_name="p",
    )


def _manual_scalar_rollout(x0: float, U: list[float], p: float, horizon: int) -> tuple[float, list[float]]:
    current = float(x0)
    packed_states = [current]
    total_cost = 0.0

    for stage_index in range(horizon):
        u_t = float(U[stage_index])
        total_cost += current * current + 2.0 * u_t * u_t + p * current * u_t
        current = 0.5 * current + p * u_t
        packed_states.append(current)

    total_cost += 3.0 * current * current + p * current
    return total_cost, packed_states


def _finite_difference_scalar_gradient(
    x0: float,
    U: list[float],
    p: float,
    horizon: int,
    *,
    epsilon: float = 1e-6,
) -> list[float]:
    gradient: list[float] = []
    for control_index in range(len(U)):
        forward = list(U)
        backward = list(U)
        forward[control_index] += epsilon
        backward[control_index] -= epsilon
        forward_cost, _ = _manual_scalar_rollout(x0, forward, p, horizon)
        backward_cost, _ = _manual_scalar_rollout(x0, backward, p, horizon)
        gradient.append((forward_cost - backward_cost) / (2.0 * epsilon))
    return gradient


class SingleShootingProblemTests(unittest.TestCase):
    def test_problem_expands_cost_gradient_and_rollout_states(self) -> None:
        problem = _build_reference_problem()
        x0 = [1.0, -0.5]
        U = [0.2, -0.1, 0.3]
        p = [0.4, -1.2]

        expected_cost, expected_states = _manual_rollout(x0, U, p, problem.horizon)
        expected_gradient = _finite_difference_gradient(x0, U, p, problem.horizon)

        cost_function = problem.to_function()
        self.assertEqual(cost_function.input_names, ("x0", "U", "p"))
        self.assertEqual(cost_function.output_names, ("cost",))
        self.assertAlmostEqual(cost_function(x0, U, p), expected_cost, places=10)

        cost_and_states = problem.to_function(include_states=True)
        actual_cost, actual_states = cost_and_states(x0, U, p)
        self.assertAlmostEqual(actual_cost, expected_cost, places=10)
        self.assertEqual(tuple(expected_states), actual_states)

        gradient_function = problem.gradient().to_function()
        actual_gradient = gradient_function(x0, U, p)
        self.assertEqual(len(actual_gradient), len(expected_gradient))
        for actual_value, expected_value in zip(actual_gradient, expected_gradient):
            self.assertAlmostEqual(actual_value, expected_value, places=5)

        gradient_and_states = problem.gradient(include_states=True).to_function()
        actual_gradient, actual_states = gradient_and_states(x0, U, p)
        self.assertEqual(tuple(expected_states), actual_states)
        for actual_value, expected_value in zip(actual_gradient, expected_gradient):
            self.assertAlmostEqual(actual_value, expected_value, places=5)

        joint_function = (
            problem.joint(
                SingleShootingBundle()
                .add_cost()
                .add_gradient()
                .add_rollout_states()
            )
            .to_function()
        )
        actual_cost, actual_gradient, actual_states = joint_function(x0, U, p)
        self.assertAlmostEqual(actual_cost, expected_cost, places=10)
        self.assertEqual(tuple(expected_states), actual_states)
        for actual_value, expected_value in zip(actual_gradient, expected_gradient):
            self.assertAlmostEqual(actual_value, expected_value, places=5)

    def test_problem_expands_hvp_and_rollout_states(self) -> None:
        # This checks the second-order path against a finite-difference
        # derivative of the control-sequence gradient, which is independent of
        # the single-shooting HVP implementation itself.
        problem = _build_reference_problem()
        x0 = [1.0, -0.5]
        U = [0.2, -0.1, 0.3]
        p = [0.4, -1.2]
        v_U = [0.5, -1.0, 0.25]

        _, expected_states = _manual_rollout(x0, U, p, problem.horizon)
        expected_hvp = _finite_difference_hvp(x0, U, p, v_U, problem.horizon)

        hvp_and_states = problem.hvp(include_states=True).to_function()
        actual_hvp, actual_states = hvp_and_states(x0, U, p, v_U)
        self.assertEqual(tuple(expected_states), actual_states)
        self.assertEqual(len(actual_hvp), len(expected_hvp))
        for actual_value, expected_value in zip(actual_hvp, expected_hvp):
            self.assertAlmostEqual(actual_value, expected_value, places=4)

    def test_problem_expands_joint_cost_gradient_hvp_and_states(self) -> None:
        # This joint expansion should agree with independent finite-difference
        # references for both the gradient and the HVP while also preserving
        # the rollout-state packing.
        problem = _build_reference_problem()
        x0 = [1.0, -0.5]
        U = [0.2, -0.1, 0.3]
        p = [0.4, -1.2]
        v_U = [0.5, -1.0, 0.25]

        expected_cost, expected_states = _manual_rollout(x0, U, p, problem.horizon)
        expected_gradient = _finite_difference_gradient(x0, U, p, problem.horizon)
        expected_hvp = _finite_difference_hvp(x0, U, p, v_U, problem.horizon)

        joint_function = (
            problem.joint(
                SingleShootingBundle()
                .add_cost()
                .add_gradient()
                .add_hvp()
                .add_rollout_states()
            )
            .to_function()
        )
        actual_cost, actual_gradient, actual_hvp, actual_states = joint_function(x0, U, p, v_U)
        self.assertAlmostEqual(actual_cost, expected_cost, places=10)
        self.assertEqual(tuple(expected_states), actual_states)
        for actual_value, expected_value in zip(actual_gradient, expected_gradient):
            self.assertAlmostEqual(actual_value, expected_value, places=5)
        for actual_value, expected_value in zip(actual_hvp, expected_hvp):
            self.assertAlmostEqual(actual_value, expected_value, places=4)

    def test_validation_rejects_mismatched_stage_signatures(self) -> None:
        x = SXVector.sym("x", 2)
        u = SXVector.sym("u", 1)
        p = SXVector.sym("p", 2)

        dynamics = Function(
            "dynamics",
            [x, u, p],
            [SXVector((x[0] + u[0], x[1] + p[0]))],
            input_names=["x", "u", "p"],
            output_names=["x_next"],
        )
        bad_stage_cost = Function(
            "bad_stage_cost",
            [x, SXVector.sym("u2", 2), p],
            [x[0] + x[1]],
            input_names=["x", "u2", "p"],
            output_names=["ell"],
        )
        terminal_cost = Function(
            "terminal_cost",
            [x, p],
            [x[0]],
            input_names=["x", "p"],
            output_names=["vf"],
        )

        with self.assertRaises(ValueError):
            SingleShootingProblem(
                name="broken",
                horizon=3,
                dynamics=dynamics,
                stage_cost=bad_stage_cost,
                terminal_cost=terminal_cost,
            )

    def test_scalar_problem_supports_scalar_state_control_and_parameter(self) -> None:
        # This exercises the SX-only single-shooting path rather than the
        # vector path. It helps catch shape bugs in packed control slicing,
        # scalar rollout flattening, and gradient expansion for horizon data.
        problem = _build_scalar_reference_problem(horizon=2)
        x0 = 1.25
        U = [0.4, -0.2]
        p = -0.75

        expected_cost, expected_states = _manual_scalar_rollout(x0, U, p, problem.horizon)
        expected_gradient = _finite_difference_scalar_gradient(x0, U, p, problem.horizon)

        self.assertEqual(problem.state_size, 1)
        self.assertEqual(problem.control_size, 1)
        self.assertEqual(problem.parameter_size, 1)

        cost_and_states = problem.to_function(include_states=True)
        actual_cost, actual_states = cost_and_states(x0, U, p)
        self.assertAlmostEqual(actual_cost, expected_cost, places=10)
        self.assertEqual(tuple(expected_states), actual_states)

        gradient_and_states = problem.gradient(include_states=True).to_function()
        actual_gradient, actual_states = gradient_and_states(x0, U, p)
        self.assertEqual(tuple(expected_states), actual_states)
        self.assertEqual(len(actual_gradient), len(expected_gradient))
        for actual_value, expected_value in zip(actual_gradient, expected_gradient):
            self.assertAlmostEqual(actual_value, expected_value, places=5)

    def test_validation_rejects_nonpositive_horizon_and_bad_terminal_output(self) -> None:
        x = SXVector.sym("x", 2)
        u = SXVector.sym("u", 1)
        p = SXVector.sym("p", 2)

        dynamics = Function(
            "dynamics",
            [x, u, p],
            [SXVector((x[0] + u[0], x[1] + p[0]))],
            input_names=["x", "u", "p"],
            output_names=["x_next"],
        )
        stage_cost = Function(
            "stage_cost",
            [x, u, p],
            [x[0] * x[0] + x[1]],
            input_names=["x", "u", "p"],
            output_names=["ell"],
        )
        bad_terminal_cost = Function(
            "bad_terminal_cost",
            [x, p],
            [SXVector((x[0], x[1]))],
            input_names=["x", "p"],
            output_names=["vf"],
        )

        with self.assertRaises(ValueError):
            SingleShootingProblem(
                name="bad_horizon",
                horizon=0,
                dynamics=dynamics,
                stage_cost=stage_cost,
                terminal_cost=Function(
                    "terminal_cost",
                    [x, p],
                    [x[0]],
                    input_names=["x", "p"],
                    output_names=["vf"],
                ),
            )

        with self.assertRaises(ValueError):
            SingleShootingProblem(
                name="bad_terminal",
                horizon=2,
                dynamics=dynamics,
                stage_cost=stage_cost,
                terminal_cost=bad_terminal_cost,
            )

    def test_joint_bundle_requires_multiple_requested_outputs(self) -> None:
        bundle = SingleShootingBundle().add_cost()
        problem = _build_reference_problem()

        with self.assertRaises(ValueError):
            problem.joint(bundle)

        with self.assertRaises(ValueError):
            problem.joint(SingleShootingBundle().add_rollout_states())

    def test_builder_accepts_single_shooting_problem(self) -> None:
        problem = _build_reference_problem()
        builder = (
            CodeGenerationBuilder()
            .with_backend_config(RustBackendConfig().with_crate_name("single_shooting"))
            .for_function(problem)
            .add_primal(include_states=True)
            .add_gradient(include_states=True)
            .add_hvp(include_states=True)
            .add_joint(
                SingleShootingBundle()
                .add_cost()
                .add_gradient()
                .add_hvp()
                .add_rollout_states()
            )
            .done()
        )

        resolved = builder._resolved_function_specs()
        self.assertEqual(len(resolved), 1)



class TestSingleShootingAdditional(unittest.TestCase):
    def test_slice_packed_sequence_scalar_and_vector(self):
        seq = SXVector.sym("U", 4)
        # scalar formal -> should return a single SX element equal to sequence[start]
        sx_formal = SX.sym("u")
        first = _slice_packed_sequence(seq, 0, 1, sx_formal)
        self.assertIsInstance(first, SX)
        self.assertEqual(first.name, seq[0].name)

        # vector formal -> returns SXVector slice of block_size
        vec_formal = SXVector.sym("uvec", 2)
        slice_vec = _slice_packed_sequence(seq, 1, 2, vec_formal)
        self.assertIsInstance(slice_vec, SXVector)
        self.assertEqual(len(slice_vec), 2)
        self.assertEqual(slice_vec[0].name, seq[2].name)

    def test_extract_single_and_scalar_output_errors_and_types(self):
        # tuple of length !=1 -> ValueError
        with self.assertRaises(ValueError):
            _extract_single_output((SX.const(1.0), SX.const(2.0)))

        # non-SX and non-tuple -> TypeError
        with self.assertRaises(TypeError):
            _extract_single_output(123)

        # scalar extractor rejects SXVector
        with self.assertRaises(ValueError):
            _extract_scalar_output(SXVector((SX.const(1.0), SX.const(2.0))))

        # valid scalar path
        self.assertEqual(_extract_scalar_output(SX.const(3.0)).value, 3.0)

    def test_flatten_rollout_states_mixes_scalar_and_vector(self):
        s1 = SX.const(1.0)
        v = SXVector((SX.const(2.0), SX.const(3.0)))
        s2 = SX.const(4.0)
        packed = _flatten_rollout_states([s1, v, s2])
        self.assertIsInstance(packed, SXVector)
        self.assertEqual(tuple(elem.value for elem in packed), (1.0, 2.0, 3.0, 4.0))

    def test_validate_single_shooting_bundle_errors_and_name_helpers(self):
        # empty bundle must raise
        empty = SingleShootingBundle()
        with self.assertRaises(ValueError):
            _validate_single_shooting_bundle(empty)

        # single flag only must raise
        only_cost = SingleShootingBundle(include_cost=True)
        with self.assertRaises(ValueError):
            _validate_single_shooting_bundle(only_cost)

        # name helpers
        self.assertEqual(_single_shooting_primal_name("base", False), "base")
        self.assertEqual(_single_shooting_primal_name("base", True), "base_with_states")

        bundle = SingleShootingBundle().add_cost().add_gradient()
        name = _single_shooting_joint_name("prob", bundle, "U")
        self.assertIn("cost", name)
        self.assertIn("gradient_U", name)
        self.assertEqual(_single_shooting_bundle_output_names(_build_minimal_scalar_problem(), bundle), ("cost", "gradient_U"))

    def test_primal_to_function_and_staged_wrappers(self):
        problem = _build_minimal_scalar_problem()
        primal = problem.primal(include_states=True, name="myp")
        func = primal.to_function()
        # inputs and outputs names should match problem
        self.assertEqual(func.input_names, problem.input_names)
        self.assertEqual(func.output_names, ("cost", "x_traj"))


if __name__ == "__main__":
    unittest.main()





"""Loop-structured single-shooting optimal-control abstractions."""

from __future__ import annotations

from dataclasses import dataclass, replace

from .function import Function
from .sx import SX, SXVector, vector


FunctionArg = SX | SXVector


@dataclass(frozen=True, slots=True)
class SingleShootingBundle:
    """Requested outputs for a joint single-shooting kernel."""

    include_cost: bool = False
    include_gradient: bool = False
    include_hvp: bool = False
    include_states: bool = False

    def add_cost(self) -> SingleShootingBundle:
        """Include the total cost in the joint kernel outputs."""
        return replace(self, include_cost=True)

    def add_gradient(self) -> SingleShootingBundle:
        """Include the gradient with respect to the packed control sequence."""
        return replace(self, include_gradient=True)

    def add_hvp(self) -> SingleShootingBundle:
        """Include the Hessian-vector product with respect to the packed control sequence."""
        return replace(self, include_hvp=True)

    def add_rollout_states(self) -> SingleShootingBundle:
        """Include the packed rollout state trajectory."""
        return replace(self, include_states=True)


@dataclass(frozen=True, slots=True)
class SingleShootingPrimalFunction:
    """Primal single-shooting cost kernel."""

    problem: SingleShootingProblem
    name: str
    include_states: bool = False
    simplification: int | str | None = None

    def to_function(self, name: str | None = None) -> Function:
        """Expand this staged kernel into a regular symbolic ``Function``."""
        function = self.problem.to_function(
            include_states=self.include_states,
            name=name or self.name,
        )
        if self.simplification is None:
            return function
        return function.simplify(max_effort=self.simplification, name=function.name)

    @property
    def nodes(self):
        """Return dependency nodes for shared helper discovery."""
        return self.to_function().nodes

    @property
    def input_names(self) -> tuple[str, ...]:
        """Return the exposed input names."""
        return self.problem.input_names

    @property
    def output_names(self) -> tuple[str, ...]:
        """Return the exposed output names."""
        names = ["cost"]
        if self.include_states:
            names.append("x_traj")
        return tuple(names)

    def generate_rust(
        self,
        *,
        config=None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Generate compact Rust for the staged primal kernel."""
        from .rust_codegen import generate_rust

        return generate_rust(
            self,
            config=config,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )

    def create_rust_project(
        self,
        path: str,
        *,
        config=None,
        crate_name: str | None = None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Create a Rust crate containing the staged primal kernel."""
        from .rust_codegen import create_rust_project

        return create_rust_project(
            self,
            path,
            config=config,
            crate_name=crate_name,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )


@dataclass(frozen=True, slots=True)
class SingleShootingGradientFunction:
    """Gradient kernel for a single-shooting optimal-control problem."""

    problem: SingleShootingProblem
    name: str
    include_states: bool = False
    simplification: int | str | None = None

    def to_function(self, name: str | None = None) -> Function:
        """Expand this staged gradient into a regular symbolic ``Function``."""
        gradient_function = self.problem._expanded_gradient_function(
            include_states=self.include_states,
            name=name or self.name,
        )
        if self.simplification is None:
            return gradient_function
        return gradient_function.simplify(
            max_effort=self.simplification,
            name=gradient_function.name,
        )

    @property
    def nodes(self):
        """Return dependency nodes for shared helper discovery."""
        return self.to_function().nodes

    @property
    def input_names(self) -> tuple[str, ...]:
        """Return the exposed input names."""
        return self.problem.input_names

    @property
    def output_names(self) -> tuple[str, ...]:
        """Return the exposed output names."""
        names = [f"gradient_{self.problem.control_sequence_name}"]
        if self.include_states:
            names.append("x_traj")
        return tuple(names)

    def generate_rust(
        self,
        *,
        config=None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Generate compact Rust for the staged gradient kernel."""
        from .rust_codegen import generate_rust

        return generate_rust(
            self,
            config=config,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )

    def create_rust_project(
        self,
        path: str,
        *,
        config=None,
        crate_name: str | None = None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Create a Rust crate containing the staged gradient kernel."""
        from .rust_codegen import create_rust_project

        return create_rust_project(
            self,
            path,
            config=config,
            crate_name=crate_name,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )


@dataclass(frozen=True, slots=True)
class SingleShootingHvpFunction:
    """HVP kernel for a single-shooting optimal-control problem."""

    problem: SingleShootingProblem
    name: str
    include_states: bool = False
    simplification: int | str | None = None

    def to_function(self, name: str | None = None) -> Function:
        """Expand this staged HVP kernel into a regular symbolic ``Function``."""
        hvp_function = self.problem._expanded_hvp_function(
            include_states=self.include_states,
            name=name or self.name,
        )
        if self.simplification is None:
            return hvp_function
        return hvp_function.simplify(
            max_effort=self.simplification,
            name=hvp_function.name,
        )

    @property
    def nodes(self):
        """Return dependency nodes for shared helper discovery."""
        return self.to_function().nodes

    @property
    def input_names(self) -> tuple[str, ...]:
        """Return the exposed input names."""
        return (*self.problem.input_names, f"v_{self.problem.control_sequence_name}")

    @property
    def output_names(self) -> tuple[str, ...]:
        """Return the exposed output names."""
        names = [f"hvp_{self.problem.control_sequence_name}"]
        if self.include_states:
            names.append("x_traj")
        return tuple(names)

    def generate_rust(
        self,
        *,
        config=None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Generate compact Rust for the staged HVP kernel."""
        from .rust_codegen import generate_rust

        return generate_rust(
            self,
            config=config,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )

    def create_rust_project(
        self,
        path: str,
        *,
        config=None,
        crate_name: str | None = None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Create a Rust crate containing the staged HVP kernel."""
        from .rust_codegen import create_rust_project

        return create_rust_project(
            self,
            path,
            config=config,
            crate_name=crate_name,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )


@dataclass(frozen=True, slots=True)
class SingleShootingJointFunction:
    """Joint cost/gradient/HVP/state kernel for a single-shooting problem."""

    problem: SingleShootingProblem
    bundle: SingleShootingBundle
    name: str
    simplification: int | str | None = None

    def to_function(self, name: str | None = None) -> Function:
        """Expand this staged joint kernel into a regular symbolic ``Function``."""
        function = self.problem._expanded_joint_function(
            self.bundle,
            name=name or self.name,
        )
        if self.simplification is None:
            return function
        return function.simplify(max_effort=self.simplification, name=function.name)

    @property
    def nodes(self):
        """Return dependency nodes for shared helper discovery."""
        return self.to_function().nodes

    @property
    def input_names(self) -> tuple[str, ...]:
        """Return the exposed input names."""
        return self.problem.input_names

    @property
    def output_names(self) -> tuple[str, ...]:
        """Return the exposed output names."""
        return _single_shooting_bundle_output_names(self.problem, self.bundle)

    def generate_rust(
        self,
        *,
        config=None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Generate compact Rust for the staged joint kernel."""
        from .rust_codegen import generate_rust

        return generate_rust(
            self,
            config=config,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )

    def create_rust_project(
        self,
        path: str,
        *,
        config=None,
        crate_name: str | None = None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Create a Rust crate containing the staged joint kernel."""
        from .rust_codegen import create_rust_project

        return create_rust_project(
            self,
            path,
            config=config,
            crate_name=crate_name,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )


@dataclass(frozen=True, slots=True)
class SingleShootingProblem:
    """Deterministic single-shooting optimal-control problem."""

    name: str
    horizon: int
    dynamics: Function
    stage_cost: Function
    terminal_cost: Function
    initial_state_name: str = "x0"
    control_sequence_name: str = "U"
    parameter_name: str = "p"
    simplification: int | str | None = None

    def __post_init__(self) -> None:
        """Validate stage and cost signatures."""
        if self.horizon <= 0:
            raise ValueError("horizon must be a positive integer")
        _validate_single_shooting_dynamics(self.dynamics)
        _validate_single_shooting_stage_cost(self.stage_cost)
        _validate_single_shooting_terminal_cost(self.terminal_cost)
        _validate_single_shooting_shapes(self.dynamics, self.stage_cost, self.terminal_cost)

    @property
    def state_size(self) -> int:
        """Return the state dimension."""
        return _single_shooting_arg_size(self.dynamics.inputs[0])

    @property
    def control_size(self) -> int:
        """Return the per-stage control dimension."""
        return _single_shooting_arg_size(self.dynamics.inputs[1])

    @property
    def parameter_size(self) -> int:
        """Return the shared parameter-vector dimension."""
        return _single_shooting_arg_size(self.dynamics.inputs[2])

    @property
    def input_names(self) -> tuple[str, ...]:
        """Return the exposed runtime input names."""
        return (
            self.initial_state_name,
            self.control_sequence_name,
            self.parameter_name,
        )

    @property
    def output_names(self) -> tuple[str, ...]:
        """Return the default primal output names."""
        return ("cost",)

    @property
    def inputs(self) -> tuple[FunctionArg, ...]:
        """Return compiled symbolic inputs."""
        return self._compiled_inputs()

    @property
    def outputs(self) -> tuple[FunctionArg, ...]:
        """Return compiled symbolic outputs for the primal cost kernel."""
        return self.to_function().outputs

    @property
    def nodes(self):
        """Return dependency nodes for shared helper discovery."""
        return self.to_function().nodes

    def primal(
        self,
        *,
        include_states: bool = False,
        name: str | None = None,
    ) -> SingleShootingProblem | SingleShootingPrimalFunction:
        """Return a staged primal kernel source."""
        if not include_states and name is None:
            return self
        return SingleShootingPrimalFunction(
            problem=self,
            name=name or _single_shooting_primal_name(self.name, include_states),
            include_states=include_states,
            simplification=self.simplification,
        )

    def gradient(
        self,
        *,
        include_states: bool = False,
        name: str | None = None,
    ) -> SingleShootingGradientFunction:
        """Return a staged gradient kernel source."""
        return SingleShootingGradientFunction(
            problem=self,
            name=name or f"{self.name}_gradient_{self.control_sequence_name}",
            include_states=include_states,
            simplification=self.simplification,
        )

    def hvp(
        self,
        *,
        include_states: bool = False,
        name: str | None = None,
    ) -> SingleShootingHvpFunction:
        """Return a staged Hessian-vector-product kernel source."""
        return SingleShootingHvpFunction(
            problem=self,
            name=name or f"{self.name}_hvp_{self.control_sequence_name}",
            include_states=include_states,
            simplification=self.simplification,
        )

    def joint(
        self,
        bundle: SingleShootingBundle,
        *,
        name: str | None = None,
    ) -> SingleShootingJointFunction:
        """Return a staged joint kernel source."""
        _validate_single_shooting_bundle(bundle)
        return SingleShootingJointFunction(
            problem=self,
            bundle=bundle,
            name=name or _single_shooting_joint_name(self.name, bundle, self.control_sequence_name),
            simplification=self.simplification,
        )

    def to_function(
        self,
        *,
        include_states: bool = False,
        name: str | None = None,
    ) -> Function:
        """Expand the total-cost kernel into a regular symbolic ``Function``."""
        x0, U, p = self._compiled_inputs()
        current_state: FunctionArg = x0
        rollout_states: list[FunctionArg] = [current_state]
        total_cost = SX.const(0.0)

        for stage_index in range(self.horizon):
            u_t = _slice_packed_sequence(U, stage_index, self.control_size, self.dynamics.inputs[1])
            total_cost = total_cost + _extract_scalar_output(
                self.stage_cost(current_state, u_t, p)
            )
            current_state = _extract_single_output(self.dynamics(current_state, u_t, p))
            rollout_states.append(current_state)

        total_cost = total_cost + _extract_scalar_output(self.terminal_cost(current_state, p))
        outputs: list[FunctionArg] = [total_cost]
        output_names = ["cost"]
        if include_states:
            outputs.append(_flatten_rollout_states(rollout_states))
            output_names.append("x_traj")

        function = Function(
            name or self.name,
            (x0, U, p),
            outputs,
            input_names=self.input_names,
            output_names=tuple(output_names),
        )
        if self.simplification is None:
            return function
        return function.simplify(max_effort=self.simplification, name=function.name)

    def _expanded_gradient_function(
        self,
        *,
        include_states: bool,
        name: str,
    ) -> Function:
        """Expand the staged gradient kernel into a symbolic ``Function``."""
        cost_function = self.to_function(include_states=False, name=f"{name}_cost")
        gradient_function = cost_function.gradient(1, name=f"{name}_grad")
        outputs: list[FunctionArg] = [gradient_function.outputs[0]]
        output_names = [f"gradient_{self.control_sequence_name}"]
        if include_states:
            outputs.append(self.to_function(include_states=True, name=f"{name}_states").outputs[1])
            output_names.append("x_traj")
        return Function(
            name,
            gradient_function.inputs,
            outputs,
            input_names=gradient_function.input_names,
            output_names=tuple(output_names),
        )

    def _expanded_hvp_function(
        self,
        *,
        include_states: bool,
        name: str,
    ) -> Function:
        """Expand the staged HVP kernel into a symbolic ``Function``."""
        cost_function = self.to_function(include_states=False, name=f"{name}_cost")
        hvp_function = cost_function.hvp(1, name=f"{name}_hvp")
        outputs: list[FunctionArg] = [hvp_function.outputs[0]]
        output_names = [f"hvp_{self.control_sequence_name}"]
        if include_states:
            outputs.append(self.to_function(include_states=True, name=f"{name}_states").outputs[1])
            output_names.append("x_traj")
        return Function(
            name,
            hvp_function.inputs,
            outputs,
            input_names=hvp_function.input_names,
            output_names=tuple(output_names),
        )

    def _expanded_joint_function(
        self,
        bundle: SingleShootingBundle,
        *,
        name: str,
    ) -> Function:
        """Expand a staged joint kernel into a symbolic ``Function``."""
        _validate_single_shooting_bundle(bundle)
        cost_function = self.to_function(include_states=False, name=f"{name}_cost")
        gradient_function = cost_function.gradient(1, name=f"{name}_grad")
        hvp_function = (
            cost_function.hvp(1, name=f"{name}_hvp")
            if bundle.include_hvp
            else None
        )
        outputs: list[FunctionArg] = []
        output_names: list[str] = []
        if bundle.include_cost:
            outputs.append(cost_function.outputs[0])
            output_names.append("cost")
        if bundle.include_gradient:
            outputs.append(gradient_function.outputs[0])
            output_names.append(f"gradient_{self.control_sequence_name}")
        if bundle.include_hvp:
            assert hvp_function is not None
            outputs.append(hvp_function.outputs[0])
            output_names.append(f"hvp_{self.control_sequence_name}")
        if bundle.include_states:
            outputs.append(self.to_function(include_states=True, name=f"{name}_states").outputs[1])
            output_names.append("x_traj")
        inputs = hvp_function.inputs if hvp_function is not None else cost_function.inputs
        input_names = hvp_function.input_names if hvp_function is not None else cost_function.input_names
        return Function(
            name,
            inputs,
            outputs,
            input_names=input_names,
            output_names=tuple(output_names),
        )

    def _compiled_inputs(self) -> tuple[FunctionArg, SXVector, FunctionArg]:
        """Return symbolic runtime inputs for ``x0``, ``U``, and ``p``."""
        x0 = _make_symbolic_like(self.dynamics.inputs[0], self.initial_state_name)
        p = _make_symbolic_like(self.dynamics.inputs[2], self.parameter_name)
        U = SXVector.sym(self.control_sequence_name, self.horizon * self.control_size)
        return x0, U, p

    def generate_rust(
        self,
        *,
        config=None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Generate compact Rust for the primal total-cost kernel."""
        from .rust_codegen import generate_rust

        return generate_rust(
            self,
            config=config,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )

    def create_rust_project(
        self,
        path: str,
        *,
        config=None,
        crate_name: str | None = None,
        function_name: str | None = None,
        backend_mode: str = "std",
        scalar_type: str = "f64",
        math_library: str | None = None,
    ):
        """Create a Rust crate containing the total-cost kernel."""
        from .rust_codegen import create_rust_project

        return create_rust_project(
            self,
            path,
            config=config,
            crate_name=crate_name,
            function_name=function_name,
            backend_mode=backend_mode,
            scalar_type=scalar_type,
            math_library=math_library,
        )


def _single_shooting_arg_size(value: FunctionArg) -> int:
    """Return the flattened scalar dimension of ``value``."""
    if isinstance(value, SX):
        return 1
    return len(value)


def _same_single_shooting_shape(left: FunctionArg, right: FunctionArg) -> bool:
    """Return ``True`` when two symbolic values share the same shape."""
    if isinstance(left, SX) and isinstance(right, SX):
        return True
    if isinstance(left, SXVector) and isinstance(right, SXVector):
        return len(left) == len(right)
    return False


def _make_symbolic_like(value: FunctionArg, base_name: str) -> FunctionArg:
    """Create a fresh symbolic input with the same shape as ``value``."""
    if isinstance(value, SX):
        return SX.sym(base_name)
    return SXVector.sym(base_name, len(value))


def _slice_packed_sequence(
    sequence: SXVector,
    stage_index: int,
    block_size: int,
    formal: FunctionArg,
) -> FunctionArg:
    """Return one stage block from a packed control sequence."""
    start = stage_index * block_size
    if isinstance(formal, SX):
        return sequence[start]
    return SXVector(sequence.elements[start : start + block_size])


def _extract_single_output(value: object) -> FunctionArg:
    """Normalize a single-output function call result."""
    if isinstance(value, tuple):
        if len(value) != 1:
            raise ValueError("expected a single function output")
        return value[0]
    if isinstance(value, (SX, SXVector)):
        return value
    raise TypeError("single-shooting stages must return symbolic outputs")


def _extract_scalar_output(value: object) -> SX:
    """Normalize a single scalar output."""
    output = _extract_single_output(value)
    if not isinstance(output, SX):
        raise ValueError("single-shooting cost functions must return scalar outputs")
    return output


def _flatten_rollout_states(states: list[FunctionArg]) -> SXVector:
    """Flatten a rollout state sequence into one packed vector."""
    scalars: list[SX] = []
    for state in states:
        if isinstance(state, SX):
            scalars.append(state)
        else:
            scalars.extend(state.elements)
    return SXVector(tuple(scalars))


def _single_shooting_primal_name(base_name: str, include_states: bool) -> str:
    """Return the default primal wrapper name."""
    if include_states:
        return f"{base_name}_with_states"
    return base_name


def _single_shooting_joint_name(
    base_name: str,
    bundle: SingleShootingBundle,
    control_sequence_name: str,
) -> str:
    """Return the default joint-kernel name."""
    labels: list[str] = []
    if bundle.include_cost:
        labels.append("cost")
    if bundle.include_gradient:
        labels.append(f"gradient_{control_sequence_name}")
    if bundle.include_hvp:
        labels.append(f"hvp_{control_sequence_name}")
    if bundle.include_states:
        labels.append("states")
    return f"{base_name}_{'_'.join(labels)}"


def _single_shooting_bundle_output_names(
    problem: SingleShootingProblem,
    bundle: SingleShootingBundle,
) -> tuple[str, ...]:
    """Return output names for a joint bundle."""
    names: list[str] = []
    if bundle.include_cost:
        names.append("cost")
    if bundle.include_gradient:
        names.append(f"gradient_{problem.control_sequence_name}")
    if bundle.include_hvp:
        names.append(f"hvp_{problem.control_sequence_name}")
    if bundle.include_states:
        names.append("x_traj")
    return tuple(names)


def _validate_single_shooting_bundle(bundle: SingleShootingBundle) -> None:
    """Validate a joint single-shooting bundle."""
    if not (bundle.include_cost or bundle.include_gradient or bundle.include_hvp):
        raise ValueError("SingleShootingBundle must request at least cost, gradient, or HVP")
    if sum((bundle.include_cost, bundle.include_gradient, bundle.include_hvp, bundle.include_states)) < 2:
        raise ValueError("joint single-shooting kernels require at least two requested outputs")


def _validate_single_shooting_dynamics(function: Function) -> None:
    """Validate the dynamics function signature."""
    if len(function.inputs) != 3:
        raise ValueError("SingleShootingProblem dynamics must accept (x, u, p)")
    if len(function.outputs) != 1:
        raise ValueError("SingleShootingProblem dynamics must return exactly one output")
    if not _same_single_shooting_shape(function.inputs[0], function.outputs[0]):
        raise ValueError("SingleShootingProblem dynamics must return the next state with the same shape as x")


def _validate_single_shooting_stage_cost(function: Function) -> None:
    """Validate the stage-cost function signature."""
    if len(function.inputs) != 3:
        raise ValueError("SingleShootingProblem stage_cost must accept (x, u, p)")
    if len(function.outputs) != 1 or not isinstance(function.outputs[0], SX):
        raise ValueError("SingleShootingProblem stage_cost must return exactly one scalar output")


def _validate_single_shooting_terminal_cost(function: Function) -> None:
    """Validate the terminal-cost function signature."""
    if len(function.inputs) != 2:
        raise ValueError("SingleShootingProblem terminal_cost must accept (x, p)")
    if len(function.outputs) != 1 or not isinstance(function.outputs[0], SX):
        raise ValueError("SingleShootingProblem terminal_cost must return exactly one scalar output")


def _validate_single_shooting_shapes(
    dynamics: Function,
    stage_cost: Function,
    terminal_cost: Function,
) -> None:
    """Validate that all stage functions agree on state, control, and parameter shapes."""
    if not _same_single_shooting_shape(dynamics.inputs[0], stage_cost.inputs[0]):
        raise ValueError("stage_cost x input must have the same shape as dynamics x")
    if not _same_single_shooting_shape(dynamics.inputs[1], stage_cost.inputs[1]):
        raise ValueError("stage_cost u input must have the same shape as dynamics u")
    if not _same_single_shooting_shape(dynamics.inputs[2], stage_cost.inputs[2]):
        raise ValueError("stage_cost p input must have the same shape as dynamics p")
    if not _same_single_shooting_shape(dynamics.outputs[0], terminal_cost.inputs[0]):
        raise ValueError("terminal_cost x input must have the same shape as the dynamics state")
    if not _same_single_shooting_shape(dynamics.inputs[2], terminal_cost.inputs[1]):
        raise ValueError("terminal_cost p input must have the same shape as dynamics p")



"""Internal helpers for Rust code generation."""

from .builder import CodeGenerationBuilder, FunctionBundle
from .naming import sanitize_ident, validate_rust_ident, validate_unique_rust_names

__all__ = [
    "CodeGenerationBuilder",
    "FunctionBundle",
    "sanitize_ident",
    "validate_rust_ident",
    "validate_unique_rust_names",
]



import unittest

from gradgen.sx import (
    SX,
    SXVector,
    acosh,
    acos,
    asinh,
    asin,
    atan2,
    atan,
    atanh,
    bilinear_form,
    cbrt,
    ceil,
    cos,
    cosh,
    erf,
    erfc,
    exp,
    expm1,
    floor,
    fract,
    hypot,
    log,
    log1p,
    matvec,
    maximum,
    minimum,
    quadform,
    round,
    signum,
    sin,
    sinh,
    sqrt,
    tan,
    tanh,
    trunc,
    vector,
)
import gradgen
from gradgen import Function


class SXTests(unittest.TestCase):
    def test_symbol_nodes_are_interned(self) -> None:
        x1 = SX.sym("x")
        x2 = SX.sym("x")

        self.assertIs(x1.node, x2.node)

    def test_symbol_metadata_participates_in_interning(self) -> None:
        x1 = SX.sym("x", metadata={"domain": "real"})
        x2 = SX.sym("x", metadata={"domain": "real"})
        xc = SX.sym("x", metadata={"domain": "complex"})

        self.assertIs(x1.node, x2.node)
        self.assertIsNot(x1.node, xc.node)

    def test_identical_binary_expressions_are_interned(self) -> None:
        x = SX.sym("x")
        y = SX.sym("y")

        expr1 = x + y
        expr2 = x + y

        self.assertIs(expr1.node, expr2.node)

    def test_commutative_nodes_are_canonicalized(self) -> None:
        x = SX.sym("x")
        y = SX.sym("y")

        self.assertIs((x + y).node, (y + x).node)
        self.assertIs((x * y).node, (y * x).node)

    def test_non_commutative_nodes_are_distinct(self) -> None:
        x = SX.sym("x")
        y = SX.sym("y")

        self.assertIsNot((x - y).node, (y - x).node)

    def test_constants_are_interned(self) -> None:
        c1 = SX.const(2.0)
        c2 = SX.const(2)

        self.assertIs(c1.node, c2.node)

    def test_unary_operations_create_expected_ops(self) -> None:
        x = SX.sym("x")

        self.assertEqual(sin(x).op, "sin")
        self.assertEqual(cos(x).op, "cos")
        self.assertEqual(tan(x).op, "tan")
        self.assertEqual(asin(x).op, "asin")
        self.assertEqual(acos(x).op, "acos")
        self.assertEqual(atan(x).op, "atan")
        self.assertEqual(asinh(x).op, "asinh")
        self.assertEqual(acosh(x).op, "acosh")
        self.assertEqual(atanh(x).op, "atanh")
        self.assertEqual(sinh(x).op, "sinh")
        self.assertEqual(cosh(x).op, "cosh")
        self.assertEqual(tanh(x).op, "tanh")
        self.assertEqual(exp(x).op, "exp")
        self.assertEqual(expm1(x).op, "expm1")
        self.assertEqual(log(x).op, "log")
        self.assertEqual(log1p(x).op, "log1p")
        self.assertEqual(sqrt(x).op, "sqrt")
        self.assertEqual(cbrt(x).op, "cbrt")
        self.assertEqual(erf(x).op, "erf")
        self.assertEqual(erfc(x).op, "erfc")
        self.assertEqual(floor(x).op, "floor")
        self.assertEqual(ceil(x).op, "ceil")
        self.assertEqual(round(x).op, "round")
        self.assertEqual(trunc(x).op, "trunc")
        self.assertEqual(fract(x).op, "fract")
        self.assertEqual(signum(x).op, "signum")
        self.assertEqual(atan2(x, 1).op, "atan2")
        self.assertEqual(hypot(x, 1).op, "hypot")
        self.assertEqual(x.abs().op, "abs")
        self.assertEqual(maximum(x, 1).op, "max")
        self.assertEqual(minimum(x, 1).op, "min")
        self.assertEqual((-x).op, "neg")

    def test_top_level_package_exports_unary_helpers(self) -> None:
        x = SX.sym("x")

        self.assertEqual(gradgen.sin(x).op, "sin")
        self.assertEqual(gradgen.cos(x).op, "cos")
        self.assertEqual(gradgen.tan(x).op, "tan")
        self.assertEqual(gradgen.asin(x).op, "asin")
        self.assertEqual(gradgen.acos(x).op, "acos")
        self.assertEqual(gradgen.atan(x).op, "atan")
        self.assertEqual(gradgen.asinh(x).op, "asinh")
        self.assertEqual(gradgen.acosh(x).op, "acosh")
        self.assertEqual(gradgen.atanh(x).op, "atanh")
        self.assertEqual(gradgen.sinh(x).op, "sinh")
        self.assertEqual(gradgen.cosh(x).op, "cosh")
        self.assertEqual(gradgen.tanh(x).op, "tanh")
        self.assertEqual(gradgen.exp(x).op, "exp")
        self.assertEqual(gradgen.expm1(x).op, "expm1")
        self.assertEqual(gradgen.log(x).op, "log")
        self.assertEqual(gradgen.log1p(x).op, "log1p")
        self.assertEqual(gradgen.sqrt(x).op, "sqrt")
        self.assertEqual(gradgen.cbrt(x).op, "cbrt")
        self.assertEqual(gradgen.erf(x).op, "erf")
        self.assertEqual(gradgen.erfc(x).op, "erfc")
        self.assertEqual(gradgen.floor(x).op, "floor")
        self.assertEqual(gradgen.ceil(x).op, "ceil")
        self.assertEqual(gradgen.round(x).op, "round")
        self.assertEqual(gradgen.trunc(x).op, "trunc")
        self.assertEqual(gradgen.fract(x).op, "fract")
        self.assertEqual(gradgen.signum(x).op, "signum")
        self.assertEqual(gradgen.atan2(x, 1).op, "atan2")
        self.assertEqual(gradgen.hypot(x, 1).op, "hypot")
        self.assertEqual(gradgen.maximum(x, 1).op, "max")
        self.assertEqual(gradgen.minimum(x, 1).op, "min")

    def test_numeric_operands_are_coerced_for_binary_operations(self) -> None:
        x = SX.sym("x")

        self.assertEqual({arg.value for arg in (x + 2).args}, {None, 2.0})
        self.assertEqual({arg.value for arg in (2 + x).args}, {None, 2.0})
        self.assertEqual({arg.value for arg in (x * 3).args}, {None, 3.0})
        self.assertEqual({arg.value for arg in (3 * x).args}, {None, 3.0})
        self.assertEqual((x / 4).args[1].value, 4.0)
        self.assertEqual((4 / x).args[0].value, 4.0)
        self.assertEqual((x**5).args[1].value, 5.0)
        self.assertEqual((5**x).args[0].value, 5.0)

    def test_reverse_and_non_commutative_binary_operators_preserve_order(self) -> None:
        x = SX.sym("x")

        expr_sub = 2 - x
        expr_div = 2 / x
        expr_pow = 2**x

        self.assertEqual(expr_sub.op, "sub")
        self.assertEqual(expr_sub.args[0].value, 2.0)
        self.assertEqual(expr_sub.args[1].name, "x")

        self.assertEqual(expr_div.op, "div")
        self.assertEqual(expr_div.args[0].value, 2.0)
        self.assertEqual(expr_div.args[1].name, "x")

        self.assertEqual(expr_pow.op, "pow")
        self.assertEqual(expr_pow.args[0].value, 2.0)
        self.assertEqual(expr_pow.args[1].name, "x")

    def test_invalid_operand_types_raise_type_error(self) -> None:
        x = SX.sym("x")

        with self.assertRaises(TypeError):
            _ = x + "bad"

        with self.assertRaises(TypeError):
            _ = sin("bad")

    def test_public_accessors_expose_node_metadata(self) -> None:
        x = SX.sym("x", metadata={"domain": "real"})
        c = SX.const(3)
        expr = x + c

        self.assertEqual(x.op, "symbol")
        self.assertEqual(x.name, "x")
        self.assertEqual(x.metadata, {"domain": "real"})
        self.assertIsNone(x.value)

        self.assertEqual(c.op, "const")
        self.assertIsNone(c.name)
        self.assertEqual(c.metadata, {})
        self.assertEqual(c.value, 3.0)

        self.assertEqual(expr.op, "add")
        self.assertEqual(expr.metadata, {})
        self.assertEqual(len(expr.args), 2)
        self.assertEqual({arg.name for arg in expr.args}, {"x", None})
        self.assertEqual({arg.value for arg in expr.args}, {None, 3.0})

    def test_repr_is_informative_for_common_node_shapes(self) -> None:
        x = SX.sym("x")
        xm = SX.sym("x", metadata={"domain": "real"})
        c = SX.const(2)

        self.assertEqual(repr(x), "SX.sym('x')")
        self.assertEqual(repr(xm), "SX.sym('x', metadata={'domain': 'real'})")
        self.assertEqual(repr(c), "SX.const(2.0)")
        self.assertEqual(repr(-x), "neg(SX.sym('x'))")
        self.assertEqual(repr(x + c), "add(SX.const(2.0), SX.sym('x'))")

    def test_symbol_metadata_is_validated(self) -> None:
        with self.assertRaises(TypeError):
            SX.sym("x", metadata="real")

        with self.assertRaises(TypeError):
            SX.sym("x", metadata={1: "real"})

        with self.assertRaises(TypeError):
            SX.sym("x", metadata={"tags": ["real"]})

    def test_nested_common_subexpressions_are_reused(self) -> None:
        x = SX.sym("x")
        y = SX.sym("y")

        subexpr = x + y
        expr = subexpr * subexpr

        self.assertIs(expr.args[0].node, expr.args[1].node)
        self.assertIs(expr.args[0].node, (x + y).node)


class SXVectorTests(unittest.TestCase):
    def test_symbolic_vector_uses_indexed_names(self) -> None:
        x = SXVector.sym("x", 3)

        self.assertEqual(len(x), 3)
        self.assertEqual([item.name for item in x], ["x_0", "x_1", "x_2"])

    def test_vector_slicing_returns_vector_view(self) -> None:
        x = SXVector.sym("x", 4)

        head = x[0:3]
        tail = x[2:]

        self.assertIsInstance(head, SXVector)
        self.assertIsInstance(tail, SXVector)
        self.assertEqual([item.name for item in head], ["x_0", "x_1", "x_2"])
        self.assertEqual([item.name for item in tail], ["x_2", "x_3"])

    def test_symbolic_vector_can_propagate_metadata(self) -> None:
        x = SXVector.sym("x", 2, metadata={"domain": "real"})

        self.assertEqual([item.metadata for item in x], [{"domain": "real"}, {"domain": "real"}])

    def test_empty_vectors_are_supported(self) -> None:
        x = SXVector.sym("x", 0)
        values = vector([])

        self.assertEqual(len(x), 0)
        self.assertEqual(len(values), 0)

    def test_empty_vector_dot_returns_zero_constant(self) -> None:
        x = SXVector.sym("x", 0)
        y = SXVector.sym("y", 0)

        dot = x.dot(y)

        self.assertEqual(dot.op, "const")
        self.assertEqual(dot.value, 0.0)

    def test_vector_constructor_coerces_scalar_like_values(self) -> None:
        values = vector([SX.sym("x"), 2, 3.5])

        self.assertEqual(len(values), 3)
        self.assertEqual(values[0].name, "x")
        self.assertEqual(values[1].value, 2.0)
        self.assertEqual(values[2].value, 3.5)

    def test_elementwise_vector_addition_and_subtraction(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)

        add_expr = x + y
        sub_expr = x - y
        reverse_sub_expr = y - x

        self.assertEqual(len(add_expr), 2)
        self.assertTrue(all(item.op == "add" for item in add_expr))
        self.assertEqual(sub_expr[0].op, "sub")
        self.assertEqual(sub_expr[0].args[0].name, "x_0")
        self.assertEqual(sub_expr[0].args[1].name, "y_0")
        self.assertEqual(reverse_sub_expr[0].args[0].name, "y_0")
        self.assertEqual(reverse_sub_expr[0].args[1].name, "x_0")

    def test_reverse_vector_addition_is_supported(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)

        expr = y.__radd__(x)

        self.assertEqual(len(expr), 2)
        self.assertTrue(all(item.op == "add" for item in expr))

    def test_scalar_vector_product_is_supported(self) -> None:
        x = SXVector.sym("x", 2)
        u = SX.sym("u")

        left = 2 * x
        right = x * 3
        symbolic_left = u * x
        symbolic_right = x * u

        self.assertTrue(all(item.op == "mul" for item in left))
        self.assertTrue(all(item.op == "mul" for item in right))
        self.assertTrue(all(item.op == "mul" for item in symbolic_left))
        self.assertTrue(all(item.op == "mul" for item in symbolic_right))
        self.assertEqual({arg.value for arg in left[0].args}, {None, 2.0})
        self.assertEqual({arg.value for arg in right[0].args}, {None, 3.0})
        self.assertEqual({arg.name for arg in symbolic_left[0].args}, {"u", "x_0"})
        self.assertEqual({arg.name for arg in symbolic_right[1].args}, {"u", "x_1"})

    def test_vector_vector_multiplication_is_not_supported(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)

        with self.assertRaises(TypeError):
            _ = x * y

        with self.assertRaises(TypeError):
            _ = x + 2

        with self.assertRaises(TypeError):
            _ = 2 + x

        with self.assertRaises(TypeError):
            _ = x / "bad"

        with self.assertRaises(TypeError):
            _ = SX.sym("s") / x

    def test_division_supports_scalar_and_vector_cases(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)

        scalar_div = x / 2
        reverse_div = 2 / x
        vector_div = x / y

        self.assertEqual(scalar_div[0].op, "div")
        self.assertEqual(scalar_div[0].args[1].value, 2.0)
        self.assertEqual(reverse_div[0].args[0].value, 2.0)
        self.assertEqual(vector_div[0].args[0].name, "x_0")
        self.assertEqual(vector_div[0].args[1].name, "y_0")

    def test_power_supports_scalar_and_vector_cases(self) -> None:
        x = SXVector.sym("x", 2)

        scalar_pow = x**2
        reverse_pow = 2**x

        self.assertEqual(scalar_pow[0].op, "pow")
        self.assertEqual(scalar_pow[0].args[1].value, 2.0)
        self.assertEqual(reverse_pow[0].op, "pow")
        self.assertEqual(reverse_pow[0].args[0].value, 2.0)

    def test_vector_power_rejects_vector_exponent(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)

        with self.assertRaises(TypeError):
            _ = x**y

        with self.assertRaises(TypeError):
            _ = y**x

    def test_vector_division_validates_lengths(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 3)

        with self.assertRaises(ValueError):
            _ = x / y

    def test_unary_vector_operations_apply_elementwise(self) -> None:
        x = SXVector.sym("x", 2)

        self.assertEqual(x.sin()[0].op, "sin")
        self.assertEqual(x.cos()[0].op, "cos")
        self.assertEqual(x.tan()[0].op, "tan")
        self.assertEqual(x.asin()[0].op, "asin")
        self.assertEqual(x.acos()[0].op, "acos")
        self.assertEqual(x.atan()[0].op, "atan")
        self.assertEqual(x.asinh()[0].op, "asinh")
        self.assertEqual(x.acosh()[0].op, "acosh")
        self.assertEqual(x.atanh()[0].op, "atanh")
        self.assertEqual(x.sinh()[0].op, "sinh")
        self.assertEqual(x.cosh()[0].op, "cosh")
        self.assertEqual(x.tanh()[0].op, "tanh")
        self.assertEqual(x.exp()[0].op, "exp")
        self.assertEqual(x.expm1()[0].op, "expm1")
        self.assertEqual(x.log()[0].op, "log")
        self.assertEqual(x.log1p()[0].op, "log1p")
        self.assertEqual(x.sqrt()[0].op, "sqrt")
        self.assertEqual(x.cbrt()[0].op, "cbrt")
        self.assertEqual(x.erf()[0].op, "erf")
        self.assertEqual(x.erfc()[0].op, "erfc")
        self.assertEqual(x.floor()[0].op, "floor")
        self.assertEqual(x.ceil()[0].op, "ceil")
        self.assertEqual(x.round()[0].op, "round")
        self.assertEqual(x.trunc()[0].op, "trunc")
        self.assertEqual(x.fract()[0].op, "fract")
        self.assertEqual(x.signum()[0].op, "signum")
        self.assertEqual(x.abs()[0].op, "abs")
        self.assertEqual((-x)[0].op, "neg")

    def test_dot_product_returns_scalar_expression(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)

        expr = x.dot(y)

        self.assertEqual(expr.op, "add")
        self.assertTrue(all(term.op == "mul" for term in expr.args))

    def test_dot_product_accepts_python_sequences(self) -> None:
        x = SXVector.sym("x", 2)
        expr = x.dot([1, 2])
        reference = x.dot(SXVector([1, 2]))

        self.assertEqual(expr.op, "add")
        self.assertIs(expr.node, reference.node)

    def test_vector_norms_build_expected_scalar_expressions(self) -> None:
        x = SXVector.sym("x", 2)

        norm1 = x.norm1()
        norm2 = x.norm2()
        norm2sq = x.norm2sq()
        norm_inf = x.norm_inf()
        norm_p = x.norm_p(3)
        norm_p_to_p = x.norm_p_to_p(3)

        self.assertEqual(norm1.op, "norm1")
        self.assertEqual(norm2.op, "norm2")
        self.assertEqual(norm2sq.op, "norm2sq")
        self.assertEqual(norm_inf.op, "norm_inf")
        self.assertEqual(norm_p.op, "norm_p")
        self.assertEqual(norm_p_to_p.op, "norm_p_to_p")

    def test_vector_reductions_build_expected_scalar_expressions(self) -> None:
        x = SXVector.sym("x", 3)

        total = x.sum()
        product = x.prod()
        maximum = x.max()
        minimum = x.min()
        mean = x.mean()

        self.assertEqual(total.op, "sum")
        self.assertEqual(product.op, "prod")
        self.assertEqual(maximum.op, "reduce_max")
        self.assertEqual(minimum.op, "reduce_min")
        self.assertEqual(mean.op, "mean")

    def test_constant_matrix_helpers_build_expected_expressions(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 2)
        matrix = [[2.0, 1.0], [1.0, 3.0]]

        matvec_expr = matvec(matrix, x)
        quadform_expr = quadform(matrix, x)
        bilinear_expr = bilinear_form(x, matrix, y)

        self.assertIsInstance(matvec_expr, SXVector)
        self.assertEqual(len(matvec_expr), 2)
        self.assertTrue(all(element.op == "matvec_component" for element in matvec_expr))
        self.assertEqual(quadform_expr.op, "quadform")
        self.assertEqual(bilinear_expr.op, "bilinear_form")

    def test_constant_matrix_helpers_validate_shapes(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 3)

        with self.assertRaises(ValueError):
            _ = matvec([[1.0, 2.0, 3.0]], x)

        with self.assertRaises(ValueError):
            _ = quadform([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], x)

        with self.assertRaises(ValueError):
            _ = quadform([[1.0, 2.0], [3.0, 4.0]], y)

        with self.assertRaises(ValueError):
            _ = bilinear_form(x, [[1.0, 2.0], [3.0, 4.0]], y)

    def test_constant_matrix_helpers_validate_matrix_entries(self) -> None:
        x = SXVector.sym("x", 2)

        with self.assertRaises(ValueError):
            _ = matvec([[1.0], [2.0, 3.0]], x)

        with self.assertRaises(TypeError):
            _ = matvec([["bad", 1.0], [2.0, 3.0]], x)

    def test_quadform_validates_symmetry_by_default(self) -> None:
        x = SXVector.sym("x", 2)

        with self.assertRaises(ValueError):
            _ = quadform([[1.0, 2.0], [3.0, 4.0]], x)

    def test_quadform_allows_non_symmetric_matrix_when_disabled(self) -> None:
        x = SXVector.sym("x", 2)
        expr = quadform([[1.0, 2.0], [3.0, 4.0]], x, is_symmetric=False)
        f = Function("quad_nonsymmetric", [x], [expr])

        # x^T P x = [2, 5] [[1,2],[3,4]] [2,5]^T = 154
        self.assertEqual(f([2.0, 5.0]), 154.0)

    def test_empty_vector_reductions_follow_documented_behavior(self) -> None:
        x = SXVector.sym("x", 0)

        self.assertEqual(x.sum().value, 0.0)
        self.assertEqual(x.prod().value, 1.0)

        with self.assertRaises(ValueError):
            _ = x.max()

        with self.assertRaises(ValueError):
            _ = x.min()

        with self.assertRaises(ValueError):
            _ = x.mean()

    def test_empty_vector_norms_return_zero_constants(self) -> None:
        x = SXVector.sym("x", 0)

        self.assertEqual(x.norm1().value, 0.0)
        self.assertEqual(x.norm2().value, 0.0)
        self.assertEqual(x.norm2sq().value, 0.0)
        self.assertEqual(x.norm_inf().value, 0.0)
        self.assertEqual(x.norm_p(3).value, 0.0)
        self.assertEqual(x.norm_p_to_p(3).value, 0.0)

    def test_vector_operations_validate_lengths(self) -> None:
        x = SXVector.sym("x", 2)
        y = SXVector.sym("y", 3)

        with self.assertRaises(ValueError):
            _ = x + y

        with self.assertRaises(ValueError):
            _ = x.dot(y)

    def test_vector_symbol_length_must_be_non_negative(self) -> None:
        with self.assertRaises(ValueError):
            SXVector.sym("x", -1)

    def test_vector_constructor_rejects_invalid_values(self) -> None:
        with self.assertRaises(TypeError):
            vector(["bad"])

    def test_scalar_vector_multiplication_reuses_canonical_nodes(self) -> None:
        x = SXVector.sym("x", 1)

        left = 2 * x
        right = x * 2

        self.assertIs(left.node, right.node)

    def test_singleton_vector_can_be_used_in_scalar_expressions(self) -> None:
        x = SXVector.sym("x", 3)
        u = SXVector.sym("u", 1)

        expr_left = u * x[0]
        expr_right = x[0] * u
        expr_add = x[0] + u

        self.assertIsInstance(expr_left, SX)
        self.assertIsInstance(expr_right, SX)
        self.assertIsInstance(expr_add, SX)
        self.assertEqual(expr_left.op, "mul")
        self.assertEqual(expr_right.op, "mul")
        self.assertEqual(expr_add.op, "add")
        self.assertEqual({arg.name for arg in expr_left.args}, {"u_0", "x_0"})

    def test_singleton_vector_supports_reverse_scalar_arithmetic(self) -> None:
        u = SXVector.sym("u", 1)

        expr_add = 2 + u
        expr_sub = 2 - u
        expr_mul = 2 * u

        self.assertIsInstance(expr_add, SX)
        self.assertIsInstance(expr_sub, SX)
        self.assertIsInstance(expr_mul, SX)
        self.assertEqual(expr_add.op, "add")
        self.assertEqual(expr_sub.op, "sub")
        self.assertEqual(expr_mul.op, "mul")
        self.assertEqual(expr_sub.args[0].value, 2.0)
        self.assertEqual(expr_sub.args[1].name, "u_0")

    def test_longer_vectors_are_still_rejected_in_scalar_expressions(self) -> None:
        x = SXVector.sym("x", 2)
        s = SX.sym("s")

        elementwise = s * x
        self.assertIsInstance(elementwise, SXVector)
        self.assertEqual(len(elementwise), 2)
        self.assertTrue(all(item.op == "mul" for item in elementwise))
        self.assertEqual({arg.name for arg in elementwise[0].args}, {"s", "x_0"})

        with self.assertRaises(TypeError):
            _ = x + s

    def test_vector_repr_is_informative(self) -> None:
        x = SXVector.sym("x", 2)

        self.assertEqual(repr(x), "SXVector(elements=(SX.sym('x_0'), SX.sym('x_1')))")

    def test_empty_matrix_helpers_support_zero_dimension_vectors(self) -> None:
        x = SXVector.sym("x", 0)
        y = SXVector.sym("y", 0)

        matvec_expr = matvec([], x)
        quadform_expr = quadform([], x)
        bilinear_expr = bilinear_form(x, [], y)

        self.assertEqual(len(matvec_expr), 0)
        self.assertEqual(quadform_expr.op, "quadform")
        self.assertEqual(bilinear_expr.op, "bilinear_form")

    def test_top_level_scalar_helpers_reject_non_singleton_vectors(self) -> None:
        x = SXVector.sym("x", 2)

        with self.assertRaises(TypeError):
            _ = maximum(x, 1)

        with self.assertRaises(TypeError):
            _ = minimum(x, 1)


if __name__ == "__main__":
    unittest.main()



import unittest

from gradgen import ComposedFunction, Function, SXVector


class ComposedFunctionTests(unittest.TestCase):
    def test_composed_function_packs_symbolic_parameters_and_evaluates(self) -> None:
        x = SXVector.sym("x", 2)
        state = SXVector.sym("state", 2)
        p = SXVector.sym("p", 2)
        pf = SXVector.sym("pf", 1)

        g = Function(
            "G",
            [state, p],
            [SXVector((state[0] + p[0], state[1] * p[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        h = Function(
            "h",
            [state, pf],
            [state[0] + state[1] + pf[0]],
            input_names=["state", "pf"],
            output_names=["y"],
        )

        composed = (
            ComposedFunction("demo", x)
            .then(g, p=p)
            .repeat(g, params=[p, p])
            .finish(h, p=pf)
        )

        compiled = composed.to_function()
        self.assertEqual(composed.input_names, ("x", "parameters"))
        self.assertEqual(composed.parameter_size, 7)
        self.assertEqual(compiled.input_names, ("x", "parameters"))
        self.assertEqual(compiled.output_names, ("y",))
        self.assertEqual(
            compiled([1.0, 2.0], [3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]),
            409.0,
        )
        self.assertEqual(
            composed.gradient().to_function()([1.0, 2.0], [3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]),
            (1.0, 192.0),
        )

    def test_repeat_requires_uniform_parameter_binding_kind(self) -> None:
        x = SXVector.sym("x", 2)
        state = SXVector.sym("state", 2)
        p = SXVector.sym("p", 2)

        g = Function(
            "G",
            [state, p],
            [SXVector((state[0] + p[0], state[1] + p[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )

        with self.assertRaises(ValueError):
            ComposedFunction("demo", x).repeat(g, params=[p, [1.0, 2.0]])

    def test_chain_accepts_function_parameter_pairs(self) -> None:
        x = SXVector.sym("x", 2)
        state = SXVector.sym("state", 2)
        p = SXVector.sym("p", 2)
        pf = SXVector.sym("pf", 1)

        g1 = Function(
            "G1",
            [state, p],
            [SXVector((state[0] + p[0], state[1] + p[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        g2 = Function(
            "G2",
            [state, p],
            [SXVector((state[0] * p[0], state[1] * p[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        h = Function(
            "h",
            [state, pf],
            [state[0] - state[1] + pf[0]],
            input_names=["state", "pf"],
            output_names=["y"],
        )

        composed = (
            ComposedFunction("demo", x)
            .chain([(g1, [1.0, 2.0]), (g2, p)])
            .finish(h, p=pf)
        )

        self.assertEqual(composed.parameter_size, 3)
        self.assertEqual(
            composed.to_function()([1.0, 3.0], [4.0, 5.0, 6.0]),
            -11.0,
        )

    def test_repeat_rejects_empty_parameter_list(self) -> None:
        x = SXVector.sym("x", 2)
        state = SXVector.sym("state", 2)
        p = SXVector.sym("p", 2)

        g = Function(
            "g",
            [state, p],
            [SXVector((state[0] + p[0], state[1] + p[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )

        with self.assertRaises(ValueError):
            ComposedFunction("demo", x).repeat(g, params=[])

    def test_stage_and_terminal_edits_are_rejected_after_finish(self) -> None:
        x = SXVector.sym("x", 2)
        state = SXVector.sym("state", 2)
        p = SXVector.sym("p", 2)
        pf = SXVector.sym("pf", 1)

        g = Function(
            "g",
            [state, p],
            [SXVector((state[0] + p[0], state[1] + p[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        h = Function(
            "h",
            [state, pf],
            [state[0] + pf[0]],
            input_names=["state", "pf"],
            output_names=["y"],
        )

        composed = ComposedFunction("demo", x).then(g, p=[1.0, 2.0]).finish(h, p=[3.0])

        with self.assertRaises(ValueError):
            composed.then(g, p=[1.0, 2.0])
        with self.assertRaises(ValueError):
            composed.repeat(g, params=[[1.0, 2.0]])
        with self.assertRaises(ValueError):
            composed.finish(h, p=[3.0])

    def test_stage_and_terminal_shape_validation_catches_dimension_mismatches(self) -> None:
        x = SXVector.sym("x", 2)
        state2 = SXVector.sym("state2", 2)
        state3 = SXVector.sym("state3", 3)
        p2 = SXVector.sym("p2", 2)
        pf = SXVector.sym("pf", 1)

        bad_stage = Function(
            "bad_stage",
            [state2, p2],
            [SXVector((state2[0], state2[1], p2[0]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        bad_terminal = Function(
            "bad_terminal",
            [state3, pf],
            [state3],
            input_names=["state", "pf"],
            output_names=["y"],
        )

        with self.assertRaises(ValueError):
            ComposedFunction("demo", x).then(bad_stage, p=[1.0, 2.0])

        good_stage = Function(
            "good_stage",
            [state2, p2],
            [SXVector((state2[0] + p2[0], state2[1] + p2[1]))],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        with self.assertRaises(ValueError):
            ComposedFunction("demo", x).then(good_stage, p=[1.0, 2.0]).finish(bad_terminal, p=[3.0])

    def test_state_input_must_be_symbolic(self) -> None:
        with self.assertRaises(ValueError):
            ComposedFunction("demo", SXVector((SXVector.sym("x", 2)[0], 1.0)))

        with self.assertRaises(ValueError):
            ComposedFunction("demo", 1.0 + SXVector.sym("x", 2)[0])


if __name__ == "__main__":
    unittest.main()



import subprocess
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import sympy as sp

from gradgen import (
    CodeGenerationBuilder,
    ComposedFunction,
    Function,
    RustBackendConfig,
    SXVector,
    SingleShootingBundle,
    SingleShootingProblem,
    map_function,
    zip_function,
)


class IntegrationTests(unittest.TestCase):
    @staticmethod
    def _run_cargo(project_dir: Path, *args: str) -> subprocess.CompletedProcess[str]:
        return subprocess.run(
            ["cargo", *args],
            cwd=project_dir,
            check=True,
            capture_output=True,
            text=True,
        )

    @staticmethod
    def _rust_array_literal(values: list[float], scalar_type: str) -> str:
        return "[" + ", ".join(f"{repr(float(value))}_{scalar_type}" for value in values) + "]"

    @staticmethod
    def _flatten_sympy_matrix(matrix: sp.Matrix) -> list[float]:
        return [float(value) for value in matrix]

    @staticmethod
    def _append_rust_test(project_dir: Path, test_source: str) -> None:
        lib_rs = project_dir / "src" / "lib.rs"
        with lib_rs.open("a", encoding="utf-8") as handle:
            handle.write("\n")
            handle.write(test_source)

    def test_scalar_function_matches_sympy_gradient_and_hessian(self) -> None:
        x_symbols = sp.symbols("x0:5", real=True)
        sympy_expr = (
            sp.exp(x_symbols[0] * x_symbols[1])
            + sp.sin(x_symbols[2]) * sp.log(1 + x_symbols[3] ** 2)
            + sp.sqrt(x_symbols[4] + 5)
            + x_symbols[0] * x_symbols[3] * x_symbols[4]
        )
        sympy_gradient = sp.Matrix([sp.diff(sympy_expr, symbol) for symbol in x_symbols])
        sympy_hessian = sp.hessian(sympy_expr, x_symbols)

        numeric_point = [0.5, -0.75, 0.3, 1.2, 2.0]
        substitutions = dict(zip(x_symbols, numeric_point))
        expected_primal = float(sympy_expr.subs(substitutions).evalf())
        expected_gradient = self._flatten_sympy_matrix(sympy_gradient.subs(substitutions).evalf())
        expected_hessian = self._flatten_sympy_matrix(sympy_hessian.subs(substitutions).evalf())

        x = SXVector.sym("x", 5)
        expr = (
            (x[0] * x[1]).exp()
            + x[2].sin() * (1 + x[3] * x[3]).log()
            + (x[4] + 5).sqrt()
            + x[0] * x[3] * x[4]
        )
        function = Function("energy", [x], [expr], input_names=["x"], output_names=["y"])

        builder = (
            CodeGenerationBuilder(function)
            .with_backend_config(RustBackendConfig().with_crate_name("sympy_scalar"))
            .add_primal()
            .add_gradient()
            .add_hessian()
            .with_simplification("medium")
        )

        with TemporaryDirectory() as tmpdir:
            project = builder.build(Path(tmpdir) / "sympy_scalar")
            primal_codegen, gradient_codegen, hessian_codegen = project.codegens

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_scalar {{
    use super::*;

    fn assert_close_slice(actual: &[f64], expected: &[f64], tolerance: f64) {{
        assert_eq!(actual.len(), expected.len());
        for (actual_value, expected_value) in actual.iter().zip(expected.iter()) {{
            assert!(
                (actual_value - expected_value).abs() <= tolerance,
                "expected {{expected_value}}, got {{actual_value}}"
            );
        }}
    }}

    #[test]
    fn matches_sympy_reference_values() {{
        let x = {self._rust_array_literal(numeric_point, "f64")};
        let mut primal_y = [0.0_f64; 1];
        let mut primal_work = [0.0_f64; {primal_codegen.workspace_size}];
        {primal_codegen.function_name}(&x, &mut primal_y, &mut primal_work);
        assert_close_slice(&primal_y, &{self._rust_array_literal([expected_primal], "f64")}, 1e-10_f64);

        let mut gradient_y = [0.0_f64; 5];
        let mut gradient_work = [0.0_f64; {gradient_codegen.workspace_size}];
        {gradient_codegen.function_name}(&x, &mut gradient_y, &mut gradient_work);
        assert_close_slice(&gradient_y, &{self._rust_array_literal(expected_gradient, "f64")}, 1e-10_f64);

        let mut hessian_y = [0.0_f64; 25];
        let mut hessian_work = [0.0_f64; {hessian_codegen.workspace_size}];
        {hessian_codegen.function_name}(&x, &mut hessian_y, &mut hessian_work);
        assert_close_slice(&hessian_y, &{self._rust_array_literal(expected_hessian, "f64")}, 1e-10_f64);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)

    def test_vector_function_matches_sympy_jacobian_and_vjp(self) -> None:
        x_symbols = sp.symbols("x0:4", real=True)
        sympy_outputs = sp.Matrix(
            [
                x_symbols[0] * x_symbols[1] + sp.sin(x_symbols[2]),
                sp.exp(x_symbols[1] - x_symbols[3]) + x_symbols[0] ** 2,
                sp.log(1 + x_symbols[0] ** 2 + x_symbols[2] ** 2) + x_symbols[3],
            ]
        )
        sympy_jacobian = sympy_outputs.jacobian(x_symbols)

        numeric_point = [0.7, -1.1, 0.4, 0.2]
        cotangent = [1.5, -0.5, 2.0]
        substitutions = dict(zip(x_symbols, numeric_point))
        expected_primal = self._flatten_sympy_matrix(sympy_outputs.subs(substitutions).evalf())
        expected_jacobian = self._flatten_sympy_matrix(sympy_jacobian.subs(substitutions).evalf())
        expected_vjp = self._flatten_sympy_matrix(
            (sympy_jacobian.subs(substitutions).T * sp.Matrix(cotangent)).evalf()
        )

        x = SXVector.sym("x", 4)
        outputs = SXVector(
            (
                x[0] * x[1] + x[2].sin(),
                (x[1] - x[3]).exp() + x[0] * x[0],
                (1 + x[0] * x[0] + x[2] * x[2]).log() + x[3],
            )
        )
        function = Function("g", [x], [outputs], input_names=["x"], output_names=["y"])

        builder = (
            CodeGenerationBuilder(function)
            .with_backend_config(RustBackendConfig().with_crate_name("sympy_vector"))
            .add_primal()
            .add_jacobian()
            .add_vjp()
            .with_simplification("medium")
        )

        with TemporaryDirectory() as tmpdir:
            project = builder.build(Path(tmpdir) / "sympy_vector")
            primal_codegen, jacobian_codegen, vjp_codegen = project.codegens

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_vector {{
    use super::*;

    fn assert_close_slice(actual: &[f64], expected: &[f64], tolerance: f64) {{
        assert_eq!(actual.len(), expected.len());
        for (actual_value, expected_value) in actual.iter().zip(expected.iter()) {{
            assert!(
                (actual_value - expected_value).abs() <= tolerance,
                "expected {{expected_value}}, got {{actual_value}}"
            );
        }}
    }}

    #[test]
    fn matches_sympy_reference_values() {{
        let x = {self._rust_array_literal(numeric_point, "f64")};
        let cotangent_y = {self._rust_array_literal(cotangent, "f64")};
        let mut primal_y = [0.0_f64; 3];
        let mut primal_work = [0.0_f64; {primal_codegen.workspace_size}];
        {primal_codegen.function_name}(&x, &mut primal_y, &mut primal_work);
        assert_close_slice(&primal_y, &{self._rust_array_literal(expected_primal, "f64")}, 1e-10_f64);

        let mut jacobian_y = [0.0_f64; 12];
        let mut jacobian_work = [0.0_f64; {jacobian_codegen.workspace_size}];
        {jacobian_codegen.function_name}(&x, &mut jacobian_y, &mut jacobian_work);
        assert_close_slice(&jacobian_y, &{self._rust_array_literal(expected_jacobian, "f64")}, 1e-10_f64);

        let mut vjp_y = [0.0_f64; 4];
        let mut vjp_work = [0.0_f64; {vjp_codegen.workspace_size}];
        {vjp_codegen.function_name}(&x, &cotangent_y, &mut vjp_y, &mut vjp_work);
        assert_close_slice(&vjp_y, &{self._rust_array_literal(expected_vjp, "f64")}, 1e-10_f64);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)

    def test_map_zip_pipeline_matches_sympy_primal_and_jacobian(self) -> None:
        # Build a SymPy reference for a 4-stage map->zip pipeline.
        # Each stage uses two state entries z=(z0,z1) and one exogenous scalar w.
        # The map stage emits a nonlinear dynamic term plus a constant term,
        # and the zip stage combines them into the final per-stage output.
        z_symbols = sp.symbols("z0:8", real=True)
        w_symbols = sp.symbols("w0:4", real=True)

        sympy_outputs: list[sp.Expr] = []
        for stage_index in range(4):
            z0 = z_symbols[2 * stage_index]
            z1 = z_symbols[(2 * stage_index) + 1]
            w = w_symbols[stage_index]
            mapped_dynamic = z0**2 + sp.sin(z1)
            mapped_constant = sp.Float("2.5")
            sympy_outputs.append(sp.exp(mapped_dynamic + w) + mapped_constant * w + 1)

        sympy_output_vector = sp.Matrix(sympy_outputs)
        sympy_jacobian = sympy_output_vector.jacobian(z_symbols)

        z_values = [0.2, -0.3, 0.4, 0.1, -0.5, 0.25, 0.3, -0.2]
        w_values = [0.05, -0.1, 0.2, -0.15]
        substitutions = {
            **{symbol: value for symbol, value in zip(z_symbols, z_values)},
            **{symbol: value for symbol, value in zip(w_symbols, w_values)},
        }
        expected_primal = self._flatten_sympy_matrix(sympy_output_vector.subs(substitutions).evalf())
        expected_jacobian = self._flatten_sympy_matrix(sympy_jacobian.subs(substitutions).evalf())

        # Symbolically define and expand the map stage over 4 stages.
        # The second output is effectively constant but still symbolic.
        z_stage = SXVector.sym("z", 2)
        map_kernel = Function(
            "map_kernel",
            [z_stage],
            [z_stage[0] * z_stage[0] + z_stage[1].sin(), z_stage[0] * 0 + 2.5],
            input_names=["z"],
            output_names=["dynamic", "constant"],
        )
        mapped = map_function(map_kernel, 4, input_name="z_seq", name="mapped_stage").to_function()

        # Define and expand the zip stage, consuming mapped outputs and w.
        dynamic_stage = SXVector.sym("dynamic", 1)
        constant_stage = SXVector.sym("constant", 1)
        w_stage = SXVector.sym("w", 1)
        zip_kernel = Function(
            "zip_kernel",
            [dynamic_stage, constant_stage, w_stage],
            [
                (dynamic_stage[0] + w_stage[0]).exp()
                + constant_stage[0] * w_stage[0]
                + 1.0
            ],
            input_names=["dynamic", "constant", "w"],
            output_names=["y"],
        )
        zipped = zip_function(
            zip_kernel,
            4,
            input_names=["dynamic_seq", "constant_seq", "w_seq"],
            name="zipped_stage",
        ).to_function()

        # Compose map + zip into one function with packed sequence inputs.
        z_seq = SXVector.sym("z_seq", 8)
        w_seq = SXVector.sym("w_seq", 4)
        mapped_dynamic_seq, mapped_constant_seq = mapped(z_seq)
        pipeline_output = zipped(mapped_dynamic_seq, mapped_constant_seq, w_seq)
        pipeline = Function(
            "map_zip_pipeline",
            [z_seq, w_seq],
            [pipeline_output],
            input_names=["z_seq", "w_seq"],
            output_names=["y_seq"],
        )

        # Generate Rust for primal and Jacobian, then verify numeric outputs
        # against the SymPy reference end-to-end through cargo test.
        builder = (
            CodeGenerationBuilder(pipeline)
            .with_backend_config(RustBackendConfig().with_crate_name("sympy_map_zip_pipeline"))
            .add_primal()
            .add_jacobian()
            .with_simplification("medium")
        )

        with TemporaryDirectory() as tmpdir:
            project = builder.build(Path(tmpdir) / "sympy_map_zip_pipeline")
            primal_codegen = next(
                codegen
                for codegen in project.codegens
                if codegen.function_name.endswith("_f")
            )
            jacobian_codegen = next(
                codegen
                for codegen in project.codegens
                if codegen.function_name.endswith("_jf_z_seq")
            )

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_map_zip_pipeline {{
    use super::*;

    fn assert_close_slice(actual: &[f64], expected: &[f64], tolerance: f64) {{
        assert_eq!(actual.len(), expected.len());
        for (actual_value, expected_value) in actual.iter().zip(expected.iter()) {{
            assert!(
                (actual_value - expected_value).abs() <= tolerance,
                "expected {{expected_value}}, got {{actual_value}}"
            );
        }}
    }}

    #[test]
    fn matches_sympy_reference_values() {{
        let z_seq = {self._rust_array_literal(z_values, "f64")};
        let w_seq = {self._rust_array_literal(w_values, "f64")};

        let mut primal_y_seq = [0.0_f64; 4];
        let mut primal_work = [0.0_f64; {primal_codegen.workspace_size}];
        {primal_codegen.function_name}(&z_seq, &w_seq, &mut primal_y_seq, &mut primal_work);
        assert_close_slice(&primal_y_seq, &{self._rust_array_literal(expected_primal, "f64")}, 1e-10_f64);

        let mut jacobian_y_seq = [0.0_f64; 32];
        let mut jacobian_work = [0.0_f64; {jacobian_codegen.workspace_size}];
        {jacobian_codegen.function_name}(&z_seq, &w_seq, &mut jacobian_y_seq, &mut jacobian_work);
        assert_close_slice(&jacobian_y_seq, &{self._rust_array_literal(expected_jacobian, "f64")}, 1e-10_f64);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)

    def test_composed_function_matches_sympy_primal_and_gradient(self) -> None:
        x0, x1 = sp.symbols("x0 x1", real=True)
        p_symbols = sp.symbols("p0:8", real=True)
        pf0, pf1 = sp.symbols("pf0 pf1", real=True)

        state = sp.Matrix([x0, x1])
        for repeat_index in range(4):
            p0 = p_symbols[2 * repeat_index]
            p1 = p_symbols[2 * repeat_index + 1]
            state = sp.Matrix(
                [
                    sp.Float("0.7") * state[0] + p0 * state[1] + sp.sin(p1),
                    sp.Float("0.2") * state[1] + p1 * state[0] + p0**2,
                ]
            )
        sympy_expr = state[0] ** 2 + pf0 * state[1] + sp.exp(state[0] - pf1)
        sympy_gradient = sp.Matrix([sp.diff(sympy_expr, x0), sp.diff(sympy_expr, x1)])

        x_values = [0.25, -0.4]
        parameter_values = [0.6, -0.2, -0.3, 0.5, 0.9, -0.7, 0.1, 0.8, 0.75, -0.1]
        substitutions = {
            x0: x_values[0],
            x1: x_values[1],
            **{symbol: value for symbol, value in zip(p_symbols, parameter_values[:8])},
            pf0: parameter_values[8],
            pf1: parameter_values[9],
        }
        expected_primal = float(sympy_expr.subs(substitutions).evalf())
        expected_gradient = self._flatten_sympy_matrix(sympy_gradient.subs(substitutions).evalf())

        x = SXVector.sym("x", 2)
        state_vector = SXVector.sym("state", 2)
        p = SXVector.sym("p", 2)
        pf = SXVector.sym("pf", 2)

        g = Function(
            "g",
            [state_vector, p],
            [
                SXVector(
                    (
                        0.7 * state_vector[0] + p[0] * state_vector[1] + p[1].sin(),
                        0.2 * state_vector[1] + p[1] * state_vector[0] + p[0] * p[0],
                    )
                )
            ],
            input_names=["state", "p"],
            output_names=["next_state"],
        )
        h = Function(
            "h",
            [state_vector, pf],
            [state_vector[0] * state_vector[0] + pf[0] * state_vector[1] + (state_vector[0] - pf[1]).exp()],
            input_names=["state", "pf"],
            output_names=["y"],
        )
        composed = ComposedFunction("sympy_composed", x).repeat(g, params=[p, p, p, p]).finish(h, p=pf)

        builder = (
            CodeGenerationBuilder()
            .with_backend_config(RustBackendConfig().with_crate_name("sympy_composed"))
            .for_function(composed)
            .add_primal()
            .add_gradient()
            .with_simplification("medium")
            .done()
        )

        with TemporaryDirectory() as tmpdir:
            project = builder.build(Path(tmpdir) / "sympy_composed")
            primal_codegen, gradient_codegen = project.codegens

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_composed {{
    use super::*;

    fn assert_close_slice(actual: &[f64], expected: &[f64], tolerance: f64) {{
        assert_eq!(actual.len(), expected.len());
        for (actual_value, expected_value) in actual.iter().zip(expected.iter()) {{
            assert!(
                (actual_value - expected_value).abs() <= tolerance,
                "expected {{expected_value}}, got {{actual_value}}"
            );
        }}
    }}

    #[test]
    fn matches_sympy_reference_values() {{
        let x = {self._rust_array_literal(x_values, "f64")};
        let parameters = {self._rust_array_literal(parameter_values, "f64")};
        let mut primal_y = [0.0_f64; 1];
        let mut primal_work = [0.0_f64; {primal_codegen.workspace_size}];
        {primal_codegen.function_name}(&x, &parameters, &mut primal_y, &mut primal_work);
        assert_close_slice(&primal_y, &{self._rust_array_literal([expected_primal], "f64")}, 1e-10_f64);

        let mut gradient_y = [0.0_f64; 2];
        let mut gradient_work = [0.0_f64; {gradient_codegen.workspace_size}];
        {gradient_codegen.function_name}(&x, &parameters, &mut gradient_y, &mut gradient_work);
        assert_close_slice(&gradient_y, &{self._rust_array_literal(expected_gradient, "f64")}, 1e-10_f64);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)

    def test_single_shooting_problem_matches_sympy_cost_gradient_and_states(self) -> None:
        sx0, sx1 = sp.symbols("x0:2", real=True)
        u_symbols = sp.symbols("u0:6", real=True)
        p0, p1 = sp.symbols("p0 p1", real=True)

        state = sp.Matrix([sx0, sx1])
        packed_states = [state[0], state[1]]
        total_cost = sp.Integer(0)
        for stage_index in range(3):
            u0 = u_symbols[2 * stage_index]
            u1 = u_symbols[(2 * stage_index) + 1]
            total_cost += (
                state[0] ** 2
                + sp.Rational(3, 5) * state[1] ** 2
                + sp.Rational(1, 5) * u0**2
                + sp.Rational(2, 5) * u1**2
                + p0 * state[0] * u1
                + p1 * u0 * state[1]
            )
            state = sp.Matrix(
                [
                    state[0] + p0 * u0 + sp.sin(u1),
                    sp.Rational(3, 5) * state[1] + p1 * u1 + state[0] * u0,
                ]
            )
            packed_states.extend([state[0], state[1]])

        total_cost += (
            sp.Rational(3, 2) * state[0] ** 2
            + sp.Rational(5, 4) * state[1] ** 2
            + p0 * state[0]
            + p1 * state[1]
        )
        gradient = sp.Matrix([sp.diff(total_cost, symbol) for symbol in u_symbols])
        sympy_states = sp.Matrix(packed_states)

        x0_values = [0.25, -0.4]
        u_values = [0.2, -0.1, 0.3, 0.15, -0.25, 0.4]
        p_values = [0.7, -0.5]
        substitutions = {
            sx0: x0_values[0],
            sx1: x0_values[1],
            **{symbol: value for symbol, value in zip(u_symbols, u_values)},
            p0: p_values[0],
            p1: p_values[1],
        }
        expected_cost = float(total_cost.subs(substitutions).evalf())
        expected_gradient = self._flatten_sympy_matrix(gradient.subs(substitutions).evalf())
        expected_states = self._flatten_sympy_matrix(sympy_states.subs(substitutions).evalf())

        x = SXVector.sym("x", 2)
        u = SXVector.sym("u", 2)
        p = SXVector.sym("p", 2)

        dynamics = Function(
            "dynamics",
            [x, u, p],
            [SXVector((x[0] + p[0] * u[0] + u[1].sin(), 0.6 * x[1] + p[1] * u[1] + x[0] * u[0]))],
            input_names=["x", "u", "p"],
            output_names=["x_next"],
        )
        stage_cost = Function(
            "stage_cost",
            [x, u, p],
            [
                x[0] * x[0]
                + 0.6 * x[1] * x[1]
                + 0.2 * u[0] * u[0]
                + 0.4 * u[1] * u[1]
                + p[0] * x[0] * u[1]
                + p[1] * u[0] * x[1]
            ],
            input_names=["x", "u", "p"],
            output_names=["ell"],
        )
        terminal_cost = Function(
            "terminal_cost",
            [x, p],
            [1.5 * x[0] * x[0] + 1.25 * x[1] * x[1] + p[0] * x[0] + p[1] * x[1]],
            input_names=["x", "p"],
            output_names=["vf"],
        )
        problem = SingleShootingProblem(
            name="sympy_single_shooting",
            horizon=3,
            dynamics=dynamics,
            stage_cost=stage_cost,
            terminal_cost=terminal_cost,
            initial_state_name="x0",
            control_sequence_name="u_seq",
            parameter_name="p",
        )

        builder = (
            CodeGenerationBuilder()
            .with_backend_config(RustBackendConfig().with_crate_name("sympy_single_shooting"))
            .for_function(problem)
            .add_joint(
                SingleShootingBundle()
                .add_cost()
                .add_gradient()
                .add_rollout_states()
            )
            .with_simplification("medium")
            .done()
        )

        with TemporaryDirectory() as tmpdir:
            project = builder.build(Path(tmpdir) / "sympy_single_shooting")
            (joint_codegen,) = project.codegens

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_single_shooting {{
    use super::*;

    fn assert_close_slice(actual: &[f64], expected: &[f64], tolerance: f64) {{
        assert_eq!(actual.len(), expected.len());
        for (actual_value, expected_value) in actual.iter().zip(expected.iter()) {{
            assert!(
                (actual_value - expected_value).abs() <= tolerance,
                "expected {{expected_value}}, got {{actual_value}}"
            );
        }}
    }}

    #[test]
    fn matches_sympy_reference_values() {{
        let x0 = {self._rust_array_literal(x0_values, "f64")};
        let u_seq = {self._rust_array_literal(u_values, "f64")};
        let p = {self._rust_array_literal(p_values, "f64")};

        let mut cost = [0.0_f64; 1];
        let mut gradient_u_seq = [0.0_f64; 6];
        let mut x_traj = [0.0_f64; 8];
        let mut work = [0.0_f64; {joint_codegen.workspace_size}];
        {joint_codegen.function_name}(&x0, &u_seq, &p, &mut cost, &mut gradient_u_seq, &mut x_traj, &mut work);

        assert_close_slice(&cost, &{self._rust_array_literal([expected_cost], "f64")}, 1e-10_f64);
        assert_close_slice(&gradient_u_seq, &{self._rust_array_literal(expected_gradient, "f64")}, 1e-10_f64);
        assert_close_slice(&x_traj, &{self._rust_array_literal(expected_states, "f64")}, 1e-10_f64);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)

    def test_single_shooting_problem_matches_sympy_cost_gradient_hvp_and_states(self) -> None:
        sx0, sx1 = sp.symbols("x0:2", real=True)
        u_symbols = sp.symbols("u0:6", real=True)
        v_symbols = sp.symbols("v0:6", real=True)
        p0, p1 = sp.symbols("p0 p1", real=True)

        state = sp.Matrix([sx0, sx1])
        packed_states = [state[0], state[1]]
        total_cost = sp.Integer(0)
        for stage_index in range(3):
            u0 = u_symbols[2 * stage_index]
            u1 = u_symbols[(2 * stage_index) + 1]
            total_cost += (
                sp.Rational(4, 5) * state[0] ** 2
                + sp.Rational(7, 10) * state[1] ** 2
                + sp.Rational(3, 10) * u0**2
                + sp.Rational(2, 5) * u1**2
                + p0 * state[0] * u1
                + p1 * state[1] * u0
                + sp.sin(u0) * state[0]
                + sp.cos(u1) * state[1]
            )
            state = sp.Matrix(
                [
                    state[0] + p0 * u0 + sp.sin(u1) + sp.Rational(1, 10) * state[1] * u0,
                    sp.Rational(11, 20) * state[1] + p1 * u1 + state[0] * u0 + sp.cos(u0),
                ]
            )
            packed_states.extend([state[0], state[1]])

        total_cost += (
            sp.Rational(9, 5) * state[0] ** 2
            + sp.Rational(6, 5) * state[1] ** 2
            + p0 * state[0]
            + p1 * state[1]
            + sp.sin(state[0] - state[1])
        )

        gradient = sp.Matrix([sp.diff(total_cost, symbol) for symbol in u_symbols])
        hessian = sp.hessian(total_cost, u_symbols)
        hvp = hessian * sp.Matrix(v_symbols)
        sympy_states = sp.Matrix(packed_states)

        x0_values = [0.1, -0.35]
        u_values = [0.15, -0.2, 0.25, 0.3, -0.1, 0.4]
        v_values = [0.5, -0.25, 0.75, -0.6, 0.2, 0.45]
        p_values = [0.6, -0.4]
        substitutions = {
            sx0: x0_values[0],
            sx1: x0_values[1],
            **{symbol: value for symbol, value in zip(u_symbols, u_values)},
            **{symbol: value for symbol, value in zip(v_symbols, v_values)},
            p0: p_values[0],
            p1: p_values[1],
        }
        expected_cost = float(total_cost.subs(substitutions).evalf())
        expected_gradient = self._flatten_sympy_matrix(gradient.subs(substitutions).evalf())
        expected_hvp = self._flatten_sympy_matrix(hvp.subs(substitutions).evalf())
        expected_states = self._flatten_sympy_matrix(sympy_states.subs(substitutions).evalf())

        x = SXVector.sym("x", 2)
        u = SXVector.sym("u", 2)
        p = SXVector.sym("p", 2)

        dynamics = Function(
            "dynamics",
            [x, u, p],
            [
                SXVector(
                    (
                        x[0] + p[0] * u[0] + u[1].sin() + 0.1 * x[1] * u[0],
                        0.55 * x[1] + p[1] * u[1] + x[0] * u[0] + u[0].cos(),
                    )
                )
            ],
            input_names=["x", "u", "p"],
            output_names=["x_next"],
        )
        stage_cost = Function(
            "stage_cost",
            [x, u, p],
            [
                0.8 * x[0] * x[0]
                + 0.7 * x[1] * x[1]
                + 0.3 * u[0] * u[0]
                + 0.4 * u[1] * u[1]
                + p[0] * x[0] * u[1]
                + p[1] * x[1] * u[0]
                + u[0].sin() * x[0]
                + u[1].cos() * x[1]
            ],
            input_names=["x", "u", "p"],
            output_names=["ell"],
        )
        terminal_cost = Function(
            "terminal_cost",
            [x, p],
            [1.8 * x[0] * x[0] + 1.2 * x[1] * x[1] + p[0] * x[0] + p[1] * x[1] + (x[0] - x[1]).sin()],
            input_names=["x", "p"],
            output_names=["vf"],
        )
        problem = SingleShootingProblem(
            name="sympy_single_shooting_joint_hvp",
            horizon=3,
            dynamics=dynamics,
            stage_cost=stage_cost,
            terminal_cost=terminal_cost,
            initial_state_name="x0",
            control_sequence_name="u_seq",
            parameter_name="p",
        )

        builder = (
            CodeGenerationBuilder()
            .with_backend_config(RustBackendConfig().with_crate_name("sympy_single_shooting_joint_hvp"))
            .for_function(problem)
            .add_joint(
                SingleShootingBundle()
                .add_cost()
                .add_gradient()
                .add_hvp()
                .add_rollout_states()
            )
            .with_simplification("medium")
            .done()
        )

        with TemporaryDirectory() as tmpdir:
            project = builder.build(Path(tmpdir) / "sympy_single_shooting_joint_hvp")
            (joint_codegen,) = project.codegens

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_single_shooting_joint_hvp {{
    use super::*;

    fn assert_close_slice(actual: &[f64], expected: &[f64], tolerance: f64) {{
        assert_eq!(actual.len(), expected.len());
        for (actual_value, expected_value) in actual.iter().zip(expected.iter()) {{
            assert!(
                (actual_value - expected_value).abs() <= tolerance,
                "expected {{expected_value}}, got {{actual_value}}"
            );
        }}
    }}

    #[test]
    fn matches_sympy_reference_values() {{
        let x0 = {self._rust_array_literal(x0_values, "f64")};
        let u_seq = {self._rust_array_literal(u_values, "f64")};
        let v_u_seq = {self._rust_array_literal(v_values, "f64")};
        let p = {self._rust_array_literal(p_values, "f64")};

        let mut cost = [0.0_f64; 1];
        let mut gradient_u_seq = [0.0_f64; 6];
        let mut hvp_u_seq = [0.0_f64; 6];
        let mut x_traj = [0.0_f64; 8];
        let mut work = [0.0_f64; {joint_codegen.workspace_size}];
        {joint_codegen.function_name}(
            &x0,
            &u_seq,
            &p,
            &v_u_seq,
            &mut cost,
            &mut gradient_u_seq,
            &mut hvp_u_seq,
            &mut x_traj,
            &mut work,
        );

        assert_close_slice(&cost, &{self._rust_array_literal([expected_cost], "f64")}, 1e-10_f64);
        assert_close_slice(&gradient_u_seq, &{self._rust_array_literal(expected_gradient, "f64")}, 1e-9_f64);
        assert_close_slice(&hvp_u_seq, &{self._rust_array_literal(expected_hvp, "f64")}, 1e-8_f64);
        assert_close_slice(&x_traj, &{self._rust_array_literal(expected_states, "f64")}, 1e-10_f64);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)



"""Public API for user-defined elementary functions.

The implementation lives in :mod:`gradgen._custom_elementary`, while this
module preserves the long-standing import surface used by the rest of the
package and by downstream users.
"""

from ._custom_elementary import (
    PythonEvalBuilder,
    RegisteredElementaryFunction,
    ScalarHessianBuilder,
    ScalarHvpBuilder,
    ScalarJacobianBuilder,
    VectorHessianBuilder,
    VectorHvpBuilder,
    VectorJacobianBuilder,
    build_custom_hessian_expr,
    build_custom_hvp_expr,
    build_custom_jacobian_expr,
    clear_registered_elementary_functions,
    custom_scalar_hessian,
    custom_scalar_hvp,
    custom_scalar_jacobian,
    custom_vector_hessian_entry,
    custom_vector_hvp_component,
    custom_vector_jacobian_component,
    evaluate_custom_hessian,
    evaluate_custom_hvp,
    evaluate_custom_jacobian,
    get_registered_elementary_function,
    parse_custom_scalar_args,
    parse_custom_scalar_hvp_args,
    parse_custom_vector_args,
    parse_custom_vector_hessian_entry_args,
    parse_custom_vector_hvp_component_args,
    parse_custom_vector_jacobian_component_args,
    register_elementary_function,
    render_custom_rust_snippet,
)
from ._custom_elementary.callbacks import (
    invoke_custom_callback as _invoke_custom_callback,
    invoke_custom_hvp_callback as _invoke_custom_hvp_callback,
)

__all__ = [
    "PythonEvalBuilder",
    "RegisteredElementaryFunction",
    "ScalarHessianBuilder",
    "ScalarHvpBuilder",
    "ScalarJacobianBuilder",
    "VectorHessianBuilder",
    "VectorHvpBuilder",
    "VectorJacobianBuilder",
    "build_custom_hessian_expr",
    "build_custom_hvp_expr",
    "build_custom_jacobian_expr",
    "clear_registered_elementary_functions",
    "custom_scalar_hessian",
    "custom_scalar_hvp",
    "custom_scalar_jacobian",
    "custom_vector_hessian_entry",
    "custom_vector_hvp_component",
    "custom_vector_jacobian_component",
    "evaluate_custom_hessian",
    "evaluate_custom_hvp",
    "evaluate_custom_jacobian",
    "get_registered_elementary_function",
    "_invoke_custom_callback",
    "_invoke_custom_hvp_callback",
    "parse_custom_scalar_args",
    "parse_custom_scalar_hvp_args",
    "parse_custom_vector_args",
    "parse_custom_vector_hessian_entry_args",
    "parse_custom_vector_hvp_component_args",
    "parse_custom_vector_jacobian_component_args",
    "register_elementary_function",
    "render_custom_rust_snippet",
]



import subprocess
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import sympy as sp

from gradgen.function import Function
from gradgen.sx import SX, SXVector
import gradgen.map_zip as map_zip


class ReduceFunctionTests(unittest.TestCase):
    @staticmethod
    def _run_cargo(project_dir: Path, *args: str) -> subprocess.CompletedProcess[str]:
        return subprocess.run(
            ["cargo", *args],
            cwd=project_dir,
            check=True,
            capture_output=True,
            text=True,
        )

    @staticmethod
    def _append_rust_test(project_dir: Path, test_source: str) -> None:
        lib_rs = project_dir / "src" / "lib.rs"
        with lib_rs.open("a", encoding="utf-8") as handle:
            handle.write("\n")
            handle.write(test_source)

    @staticmethod
    def _rust_array_literal(values: list[float], scalar_type: str) -> str:
        return "[" + ", ".join(f"{repr(float(value))}_{scalar_type}" for value in values) + "]"

    def _require_reduce_api(self):
        self.assertTrue(
            hasattr(map_zip, "reduce_function"),
            "Expected gradgen.map_zip.reduce_function to exist",
        )
        self.assertTrue(
            hasattr(map_zip, "ReducedFunction"),
            "Expected gradgen.map_zip.ReducedFunction to exist",
        )
        return map_zip.reduce_function, map_zip.ReducedFunction

    def test_reduce_api_exists(self) -> None:
        self._require_reduce_api()

    def test_reduce_validates_stage_function_signature(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        x = SX.sym("x")
        y = SX.sym("y")
        z = SX.sym("z")

        unary = Function("unary", [x], [x + 1.0])
        ternary = Function("ternary", [x, y, z], [x + y + z])
        two_outputs = Function("two_outputs", [x, y], [x + y, x - y])

        with self.assertRaises(ValueError):
            reduce_function(unary, 3)
        with self.assertRaises(ValueError):
            reduce_function(ternary, 3)
        with self.assertRaises(ValueError):
            reduce_function(two_outputs, 3)

    def test_reduce_requires_output_shape_to_match_accumulator_shape(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        acc = SXVector.sym("acc", 2)
        x = SX.sym("x")
        bad_stage = Function(
            "bad_stage",
            [acc, x],
            [x],
            input_names=["acc", "x"],
            output_names=["acc_next"],
        )

        with self.assertRaises(ValueError):
            reduce_function(bad_stage, 2)

    def test_reduce_to_function_matches_left_fold_scalar(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        acc = SX.sym("acc")
        x = SX.sym("x")
        stage = Function(
            "stage",
            [acc, x],
            [acc + 2.0 * x],
            input_names=["acc", "x"],
            output_names=["acc_next"],
        )

        reduced = reduce_function(
            stage,
            4,
            accumulator_input_name="acc0",
            input_name="x_seq",
            name="stage_reduce",
            simplification="medium",
        )
        expanded = reduced.to_function()

        # (((1 + 2*3) + 2*(-1)) + 2*4) + 2*0.5 = 14
        result = expanded(1.0, [3.0, -1.0, 4.0, 0.5])
        self.assertAlmostEqual(result, 14.0)

    def test_reduce_respects_left_fold_order_for_non_commutative_stage(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        acc = SX.sym("acc")
        x = SX.sym("x")
        stage = Function(
            "stage_non_commutative",
            [acc, x],
            [acc - x],
            input_names=["acc", "x"],
            output_names=["acc_next"],
        )

        reduced = reduce_function(stage, 3, accumulator_input_name="acc0", input_name="x_seq")
        expanded = reduced.to_function()

        self.assertAlmostEqual(expanded(10.0, [1.0, 2.0, 3.0]), 4.0)

    def test_reduce_supports_vector_accumulator_and_vector_sequence_element(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        acc = SXVector.sym("acc", 2)
        x = SXVector.sym("x", 2)
        stage = Function(
            "vec_stage",
            [acc, x],
            [SXVector((acc[0] + x[0], acc[1] - x[1]))],
            input_names=["acc", "x"],
            output_names=["acc_next"],
        )

        reduced = reduce_function(
            stage,
            3,
            accumulator_input_name="acc0",
            input_name="x_seq",
            name="vec_reduce",
        )
        expanded = reduced.to_function()

        # acc0=(1,10), x_seq=[(2,3), (-1,4), (5,-2)]
        # step1 -> (3,7), step2 -> (2,3), step3 -> (7,5)
        self.assertEqual(expanded([1.0, 10.0], [2.0, 3.0, -1.0, 4.0, 5.0, -2.0]), (7.0, 5.0))

    def test_reduce_codegen_preserves_loop_structure(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        acc = SX.sym("acc")
        x = SX.sym("x")
        stage = Function("stage", [acc, x], [acc + x], input_names=["acc", "x"], output_names=["acc_next"])
        reduced = reduce_function(stage, 5, accumulator_input_name="acc0", input_name="x_seq", name="sum_reduce")

        codegen = reduced.generate_rust(function_name="sum_reduce_kernel")

        self.assertIn("for stage_index in 0..5", codegen.source)

    def test_reduce_generated_rust_matches_sympy_reference(self) -> None:
        reduce_function, _ = self._require_reduce_api()

        count = 4
        acc0_symbol = sp.Symbol("acc0", real=True)
        x_symbols = sp.symbols("x0:4", real=True)

        sympy_acc = acc0_symbol
        for x_i in x_symbols:
            sympy_acc = sp.sin(sympy_acc) + (sympy_acc * x_i) + (x_i**2)

        acc0_value = 0.35
        x_values = [0.2, -0.4, 0.6, 0.1]
        substitutions = {acc0_symbol: acc0_value, **{sym: val for sym, val in zip(x_symbols, x_values)}}
        expected = float(sympy_acc.subs(substitutions).evalf())

        acc = SX.sym("acc")
        x = SX.sym("x")
        stage = Function(
            "nonlinear_stage",
            [acc, x],
            [acc.sin() + acc * x + x * x],
            input_names=["acc", "x"],
            output_names=["acc_next"],
        )
        reduced = reduce_function(
            stage,
            count,
            accumulator_input_name="acc0",
            input_name="x_seq",
            name="nonlinear_reduce",
            simplification="medium",
        )

        with TemporaryDirectory() as tmpdir:
            project = reduced.create_rust_project(Path(tmpdir) / "nonlinear_reduce_kernel")

            self._append_rust_test(
                project.project_dir,
                f"""
#[cfg(test)]
mod integration_sympy_reduce {{
    use super::*;

    #[test]
    fn generated_reduce_matches_sympy_reference() {{
        let acc0 = {self._rust_array_literal([acc0_value], "f64")};
        let x_seq = {self._rust_array_literal(x_values, "f64")};
        let expected = {self._rust_array_literal([expected], "f64")};

        let mut out = [0.0_f64; 1];
        let mut work = [0.0_f64; {project.codegen.workspace_size}];
        {project.codegen.function_name}(&acc0, &x_seq, &mut out, &mut work)
            .expect("reduce kernel evaluation failed");

        assert!((out[0] - expected[0]).abs() <= 1e-10_f64, "expected {{}} got {{}}", expected[0], out[0]);
    }}
}}
""".lstrip(),
            )

            completed = self._run_cargo(project.project_dir, "test", "--quiet")
            self.assertEqual(completed.returncode, 0)


if __name__ == "__main__":
    unittest.main()



"""Builder API for creating multi-kernel Rust crates."""

from __future__ import annotations

from dataclasses import dataclass, replace
from pathlib import Path
from typing import Callable, Protocol

from ..composed_function import ComposedFunction, ComposedGradientFunction
from ..function import Function
from ..map_zip import ZippedFunction, ZippedJacobianFunction, ReducedFunction
from ..single_shooting import (
    SingleShootingBundle,
    SingleShootingGradientFunction,
    SingleShootingHvpFunction,
    SingleShootingJointFunction,
    SingleShootingPrimalFunction,
    SingleShootingProblem,
)
from .naming import sanitize_ident


BuilderSource = (
    Function
    | ComposedFunction
    | ComposedGradientFunction
    | ZippedFunction
    | ZippedJacobianFunction
    | ReducedFunction
    | SingleShootingProblem
    | SingleShootingPrimalFunction
    | SingleShootingGradientFunction
    | SingleShootingHvpFunction
    | SingleShootingJointFunction
)


class RustBuilderConfigLike(Protocol):
    """Protocol describing the config surface used by the builder."""

    crate_name: str | None
    function_name: str | None

    def with_crate_name(self, crate_name: str | None) -> RustBuilderConfigLike:
        """Return a copy with an updated crate name."""


@dataclass(frozen=True, slots=True)
class _BuilderRequest:
    """A requested generated kernel kind."""

    kind: str
    components: tuple[str, ...] = ()
    bundle: FunctionBundle | SingleShootingBundle | None = None


@dataclass(frozen=True, slots=True)
class _BundleItem:
    """One artifact entry inside a ``FunctionBundle``."""

    kind: str
    wrt_indices: tuple[int, ...] | None = None


@dataclass(frozen=True, slots=True)
class FunctionBundle:
    """Fluent description of artifacts to compute together in one joint kernel."""

    items: tuple[_BundleItem, ...] = ()

    def add_f(self) -> FunctionBundle:
        """Include the primal outputs."""
        return self._add_item("f")

    def add_gradient(self, *, wrt: int | list[int] | tuple[int, ...]) -> FunctionBundle:
        """Include gradient blocks for one or more input indices."""
        return self._add_item("grad", wrt)

    def add_jf(self, *, wrt: int | list[int] | tuple[int, ...]) -> FunctionBundle:
        """Include Jacobian blocks for one or more input indices."""
        return self._add_item("jf", wrt)

    def add_hessian(self, *, wrt: int | list[int] | tuple[int, ...]) -> FunctionBundle:
        """Include Hessian blocks for one or more input indices."""
        return self._add_item("hessian", wrt)

    def add_hvp(self, *, wrt: int | list[int] | tuple[int, ...]) -> FunctionBundle:
        """Include Hessian-vector-product blocks for one or more input indices."""
        return self._add_item("hvp", wrt)

    def _add_item(
        self,
        kind: str,
        wrt: int | list[int] | tuple[int, ...] | None = None,
    ) -> FunctionBundle:
        wrt_indices = None if wrt is None else _normalize_wrt_indices(wrt)
        candidate = _BundleItem(kind, wrt_indices)
        if any(item == candidate for item in self.items):
            return self
        return replace(self, items=(*self.items, candidate))


@dataclass(frozen=True, slots=True)
class _BuilderFunctionSpec:
    """One source function plus the kernels requested for it."""

    function: BuilderSource
    requests: tuple[_BuilderRequest, ...] = ()
    simplification: int | str | None = None


@dataclass(frozen=True, slots=True)
class CodeGenerationBuilder:
    """Fluent builder for a generated Rust crate with one or more functions."""

    function: BuilderSource | None = None
    config: RustBuilderConfigLike | None = None
    requests: tuple[_BuilderRequest, ...] = ()
    simplification: int | str | None = None
    functions: tuple[_BuilderFunctionSpec, ...] = ()

    def with_backend_config(self, config: RustBuilderConfigLike) -> CodeGenerationBuilder:
        """Return a copy using ``config`` for generated Rust code."""
        return replace(self, config=config)

    def with_simplification(self, max_effort: int | str | None) -> CodeGenerationBuilder:
        """Return a copy applying ``max_effort`` simplification to generated kernels."""
        return replace(self, simplification=max_effort)

    def add_primal(self, *, include_states: bool = False) -> CodeGenerationBuilder:
        """Include the primal function in the generated crate."""
        components = ("states",) if include_states else ()
        return self._add_request("primal", components=components)

    def add_gradient(self, *, include_states: bool = False) -> CodeGenerationBuilder:
        """Include gradient kernels for scalar-output functions."""
        components = ("states",) if include_states else ()
        return self._add_request("gradient", components=components)

    def add_jacobian(self) -> CodeGenerationBuilder:
        """Include Jacobian kernels for all input blocks."""
        return self._add_request("jacobian")

    def add_vjp(self) -> CodeGenerationBuilder:
        """Include runtime-seeded vector-Jacobian-product kernels for input blocks."""
        return self._add_request("vjp")

    def add_joint(self, bundle: FunctionBundle | SingleShootingBundle) -> CodeGenerationBuilder:
        """Include kernels that compute bundled artifacts together."""
        return self._add_request("joint", bundle=bundle)

    def add_hessian(self) -> CodeGenerationBuilder:
        """Include Hessian kernels for scalar-output functions."""
        return self._add_request("hessian")

    def add_hvp(self, *, include_states: bool = False) -> CodeGenerationBuilder:
        """Include Hessian-vector product kernels for scalar-output functions."""
        components = ("states",) if include_states else ()
        return self._add_request("hvp", components=components)

    def for_function(
        self,
        function: BuilderSource,
        configure: Callable[[CodeGenerationBuilder], CodeGenerationBuilder] | None = None,
    ) -> CodeGenerationBuilder | FunctionCodegenBuilder:
        """Start configuring kernels for ``function`` or apply a legacy callback."""
        scoped_builder = FunctionCodegenBuilder(parent=self, function=function)
        if configure is None:
            return scoped_builder
        configured_builder = configure(
            CodeGenerationBuilder(function=function, config=self.config)
        )
        return _append_builder_function_spec(
            self,
            _BuilderFunctionSpec(
                function=_require_builder_function(configured_builder),
                requests=configured_builder.requests,
                simplification=configured_builder.simplification,
            ),
        )

    def build(self, path: str | Path):
        """Generate a single Rust crate containing all requested kernels."""
        from ..rust_codegen import RustBackendConfig, create_multi_function_rust_project

        resolved_config = self.config or RustBackendConfig()
        if resolved_config.crate_name is None:
            resolved_config = resolved_config.with_crate_name(
                sanitize_ident(Path(path).expanduser().resolve().name)
            )
        functions = resolve_builder_function_specs(
            self._resolved_function_specs(),
            resolved_config,
        )
        return create_multi_function_rust_project(
            functions,
            path,
            config=resolved_config,
        )

    def _add_request(
        self,
        kind: str,
        *,
        components: tuple[str, ...] = (),
        bundle: FunctionBundle | None = None,
    ) -> CodeGenerationBuilder:
        if self.function is None:
            raise ValueError(
                "no source function is selected; initialize CodeGenerationBuilder(function) "
                "or use for_function(...)"
            )
        candidate = _BuilderRequest(kind, components, bundle)
        if any(request == candidate for request in self.requests):
            return self
        return replace(self, requests=(*self.requests, candidate))

    def _resolved_function_specs(self) -> tuple[_BuilderFunctionSpec, ...]:
        """Return all source-function entries tracked by the builder."""
        root_specs: tuple[_BuilderFunctionSpec, ...] = ()
        if self.function is not None and self.requests:
            root_specs = (
                _BuilderFunctionSpec(
                    function=self.function,
                    requests=self.requests,
                    simplification=self.simplification,
                ),
            )
        specs = (*root_specs, *self.functions)
        if not specs:
            raise ValueError(
                "no kernels were requested; call add_primal() or another add_* method first"
            )
        return specs


@dataclass(frozen=True, slots=True)
class FunctionCodegenBuilder:
    """Scoped builder used to configure kernels for one source function."""

    parent: CodeGenerationBuilder
    function: BuilderSource
    requests: tuple[_BuilderRequest, ...] = ()
    simplification: int | str | None = None

    def with_simplification(self, max_effort: int | str | None) -> FunctionCodegenBuilder:
        """Return a copy applying ``max_effort`` simplification to this function."""
        return replace(self, simplification=max_effort)

    def add_primal(self, *, include_states: bool = False) -> FunctionCodegenBuilder:
        """Include the primal function in the generated crate."""
        components = ("states",) if include_states else ()
        return self._add_request("primal", components=components)

    def add_gradient(self, *, include_states: bool = False) -> FunctionCodegenBuilder:
        """Include gradient kernels for scalar-output functions."""
        components = ("states",) if include_states else ()
        return self._add_request("gradient", components=components)

    def add_jacobian(self) -> FunctionCodegenBuilder:
        """Include Jacobian kernels for all input blocks."""
        return self._add_request("jacobian")

    def add_vjp(self) -> FunctionCodegenBuilder:
        """Include runtime-seeded vector-Jacobian-product kernels for input blocks."""
        return self._add_request("vjp")

    def add_joint(self, bundle: FunctionBundle | SingleShootingBundle) -> FunctionCodegenBuilder:
        """Include kernels that compute bundled artifacts together."""
        return self._add_request("joint", bundle=bundle)

    def add_hessian(self) -> FunctionCodegenBuilder:
        """Include Hessian kernels for scalar-output functions."""
        return self._add_request("hessian")

    def add_hvp(self, *, include_states: bool = False) -> FunctionCodegenBuilder:
        """Include Hessian-vector product kernels for scalar-output functions."""
        components = ("states",) if include_states else ()
        return self._add_request("hvp", components=components)

    def done(self) -> CodeGenerationBuilder:
        """Commit this scoped function configuration back to the parent builder."""
        if not self.requests:
            raise ValueError(
                "no kernels were requested for this function; call add_primal() or another add_* method first"
            )
        return _append_builder_function_spec(
            self.parent,
            _BuilderFunctionSpec(
                function=self.function,
                requests=self.requests,
                simplification=self.simplification,
            ),
        )

    def _add_request(
        self,
        kind: str,
        *,
        components: tuple[str, ...] = (),
        bundle: FunctionBundle | None = None,
    ) -> FunctionCodegenBuilder:
        candidate = _BuilderRequest(kind, components, bundle)
        if any(request == candidate for request in self.requests):
            return self
        return replace(self, requests=(*self.requests, candidate))


def _append_builder_function_spec(
    builder: CodeGenerationBuilder,
    spec: _BuilderFunctionSpec,
) -> CodeGenerationBuilder:
    """Return ``builder`` with one resolved function specification appended."""
    return replace(builder, functions=(*builder.functions, spec))


def _require_builder_function(builder: CodeGenerationBuilder) -> BuilderSource:
    """Return the selected source function or raise for misconfigured callback builders."""
    if builder.function is None:
        raise ValueError("configured function builder must target a function")
    return builder.function


def resolve_builder_function_specs(
    specs: tuple[_BuilderFunctionSpec, ...],
    config: RustBuilderConfigLike,
) -> tuple[BuilderSource, ...]:
    """Expand builder function specs into concrete symbolic functions."""
    if len(specs) > 1 and config.function_name is not None:
        raise ValueError(
            "RustBackendConfig.function_name is only supported when the builder targets "
            "a single source function"
        )

    include_base_name = True
    resolved: list[Function] = []
    for spec in specs:
        base_name = config.function_name if len(specs) == 1 else None
        resolved.extend(
            _resolve_builder_functions(
                spec.function,
                config,
                spec.requests,
                spec.simplification,
                include_base_name=include_base_name,
                function_name=base_name,
            )
        )
    return tuple(resolved)


def _resolve_builder_functions(
    function: BuilderSource,
    config: RustBuilderConfigLike,
    requests: tuple[_BuilderRequest, ...],
    simplification: int | str | None,
    *,
    include_base_name: bool = False,
    function_name: str | None = None,
) -> tuple[BuilderSource, ...]:
    """Expand builder requests into concrete symbolic functions."""
    if not requests:
        raise ValueError("no kernels were requested; call add_primal() or another add_* method first")

    if isinstance(function, (ComposedFunction, ComposedGradientFunction)):
        return _resolve_builder_composed_sources(
            function,
            config,
            requests,
            simplification,
            include_base_name=include_base_name,
            function_name=function_name,
        )
    if isinstance(function, (ZippedFunction, ZippedJacobianFunction, ReducedFunction)):
        return _resolve_builder_zipped_sources(
            function,
            config,
            requests,
            simplification,
            include_base_name=include_base_name,
            function_name=function_name,
        )
    if isinstance(
        function,
        (
            SingleShootingProblem,
            SingleShootingPrimalFunction,
            SingleShootingGradientFunction,
            SingleShootingHvpFunction,
            SingleShootingJointFunction,
        ),
    ):
        return _resolve_builder_single_shooting_sources(
            function,
            config,
            requests,
            simplification,
            include_base_name=include_base_name,
            function_name=function_name,
        )

    base_function = _apply_builder_base_name(function, function_name)
    crate_prefix = sanitize_ident(config.crate_name or base_function.name)
    resolved: list[Function] = []

    for request in requests:
        if request.kind == "primal":
            if request.components:
                raise ValueError("include_states is only supported for SingleShootingProblem sources")
            resolved.append(
                _rename_generated_function(
                    _maybe_simplify_generated_function(base_function, simplification),
                    _builder_function_name(
                        crate_prefix,
                        "f",
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                    ),
                )
            )
            continue
        if request.kind == "gradient":
            if request.components:
                raise ValueError("include_states is only supported for SingleShootingProblem sources")
            resolved.extend(
                _rename_generated_function(
                    _maybe_simplify_generated_function(base_function.gradient(index), simplification),
                    _builder_function_name(
                        crate_prefix,
                        "grad",
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                        input_name=base_function.input_names[index],
                        include_input_name=len(base_function.inputs) > 1,
                    ),
                )
                for index in range(len(base_function.inputs))
            )
            continue
        if request.kind == "jacobian":
            resolved.extend(
                _rename_generated_function(
                    _maybe_simplify_generated_function(block, simplification),
                    _builder_function_name(
                        crate_prefix,
                        "jf",
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                        input_name=base_function.input_names[index],
                        include_input_name=len(base_function.inputs) > 1,
                    ),
                )
                for index, block in enumerate(base_function.jacobian_blocks())
            )
            continue
        if request.kind == "vjp":
            resolved.extend(
                _rename_generated_function(
                    _maybe_simplify_generated_function(block, simplification),
                    _builder_function_name(
                        crate_prefix,
                        "vjp",
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                        input_name=base_function.input_names[index],
                        include_input_name=len(base_function.inputs) > 1,
                    ),
                )
                for index, block in enumerate(base_function.vjp_blocks())
            )
            continue
        if request.kind == "joint":
            if request.bundle is None:
                raise ValueError("joint builder requests require a FunctionBundle")
            resolved.extend(
                _rename_generated_function(
                    _maybe_simplify_generated_function(
                        base_function.joint(components, index),
                        simplification,
                    ),
                    _builder_function_name(
                        crate_prefix,
                        *_builder_joint_labels(components),
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                        input_name=base_function.input_names[index],
                        include_input_name=len(base_function.inputs) > 1,
                    ),
                )
                for index, components in _resolve_function_bundle(
                    request.bundle,
                    len(base_function.inputs),
                )
            )
            continue
        if request.kind == "hessian":
            resolved.extend(
                _rename_generated_function(
                    _maybe_simplify_generated_function(block, simplification),
                    _builder_function_name(
                        crate_prefix,
                        "hessian",
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                        input_name=base_function.input_names[index],
                        include_input_name=len(base_function.inputs) > 1,
                    ),
                )
                for index, block in enumerate(base_function.hessian_blocks())
            )
            continue
        if request.kind == "hvp":
            if request.components:
                raise ValueError("include_states is only supported for SingleShootingProblem sources")
            resolved.extend(
                _rename_generated_function(
                    _maybe_simplify_generated_function(block, simplification),
                    _builder_function_name(
                        crate_prefix,
                        "hvp",
                        base_name=base_function.name,
                        include_base_name=include_base_name,
                        input_name=base_function.input_names[index],
                        include_input_name=len(base_function.inputs) > 1,
                    ),
                )
                for index, block in enumerate(base_function.hvp_blocks())
            )
            continue
        raise ValueError(f"unsupported builder request kind {request.kind!r}")

    return tuple(resolved)


def _resolve_builder_composed_sources(
    function: ComposedFunction | ComposedGradientFunction,
    config: RustBuilderConfigLike,
    requests: tuple[_BuilderRequest, ...],
    simplification: int | str | None,
    *,
    include_base_name: bool = False,
    function_name: str | None = None,
) -> tuple[BuilderSource, ...]:
    """Expand builder requests for staged composed sources."""
    crate_prefix = sanitize_ident(config.crate_name or function.name)
    base_name = function_name or function.name
    resolved: list[BuilderSource] = []

    for request in requests:
        if request.kind == "primal":
            if request.components:
                raise ValueError("include_states is only supported for SingleShootingProblem sources")
            resolved.append(
                _rename_builder_source(
                    replace(function, simplification=simplification),
                    _builder_function_name(
                        crate_prefix,
                        "f",
                        base_name=base_name,
                        include_base_name=include_base_name,
                    ),
                )
            )
            continue
        if request.kind == "gradient" and isinstance(function, ComposedFunction):
            if request.components:
                raise ValueError("include_states is only supported for SingleShootingProblem sources")
            simplified_function = replace(function, simplification=simplification)
            resolved.append(
                _rename_builder_source(
                    simplified_function.gradient(),
                    _builder_function_name(
                        crate_prefix,
                        "grad",
                        base_name=base_name,
                        include_base_name=include_base_name,
                        input_name=function.input_name,
                        include_input_name=True,
                    ),
                )
            )
            continue
        raise ValueError(
            "staged composed sources currently support only add_primal() and "
            "ComposedFunction.add_gradient() in CodeGenerationBuilder"
        )

    return tuple(resolved)


def _resolve_builder_zipped_sources(
    function: ZippedFunction | ZippedJacobianFunction | ReducedFunction,
    config: RustBuilderConfigLike,
    requests: tuple[_BuilderRequest, ...],
    simplification: int | str | None,
    *,
    include_base_name: bool = False,
    function_name: str | None = None,
) -> tuple[BuilderSource, ...]:
    """Expand builder requests for staged map/zip/reduce sources."""
    crate_prefix = sanitize_ident(config.crate_name or function.name)
    base_name = function_name or function.name
    resolved: list[BuilderSource] = []

    if isinstance(function, ZippedJacobianFunction):
        for request in requests:
            if request.kind != "primal":
                raise ValueError(
                    "explicit staged Jacobian map/zip sources in CodeGenerationBuilder "
                    "currently support only add_primal()"
                )
            if request.components:
                raise ValueError("include_states is not supported for map/zip/reduce sources")
            resolved.append(
                _rename_builder_source(
                    replace(function, simplification=simplification),
                    _builder_function_name(
                        crate_prefix,
                        "f",
                        base_name=base_name,
                        include_base_name=include_base_name,
                    ),
                )
            )
        return tuple(resolved)

    simplified_function = replace(function, simplification=simplification)
    for request in requests:
        if request.kind == "primal":
            if request.components:
                raise ValueError("include_states is not supported for map/zip/reduce sources")
            resolved.append(
                _rename_builder_source(
                    simplified_function,
                    _builder_function_name(
                        crate_prefix,
                        "f",
                        base_name=base_name,
                        include_base_name=include_base_name,
                    ),
                )
            )
            continue
        if request.kind == "jacobian":
            if request.components:
                raise ValueError("include_states is not supported for map/zip/reduce sources")
            if not isinstance(simplified_function, ZippedFunction):
                raise ValueError("add_jacobian() is only supported for map/zip sources")
            for input_index, input_name in enumerate(simplified_function.input_names):
                resolved.append(
                    _rename_builder_source(
                        simplified_function.jacobian(input_index),
                        _builder_function_name(
                            crate_prefix,
                            "jf",
                            base_name=base_name,
                            include_base_name=include_base_name,
                            input_name=input_name,
                            include_input_name=True,
                        ),
                    )
                )
            continue
        raise ValueError(
            "map/zip/reduce sources currently support only add_primal(), and add_jacobian() only for map/zip"
        )

    return tuple(resolved)


def _resolve_builder_single_shooting_sources(
    function: (
        SingleShootingProblem
        | SingleShootingPrimalFunction
        | SingleShootingGradientFunction
        | SingleShootingJointFunction
    ),
    config: RustBuilderConfigLike,
    requests: tuple[_BuilderRequest, ...],
    simplification: int | str | None,
    *,
    include_base_name: bool = False,
    function_name: str | None = None,
) -> tuple[BuilderSource, ...]:
    """Expand builder requests for single-shooting sources."""
    crate_prefix = sanitize_ident(config.crate_name or function.name)
    base_name = function_name or function.name
    resolved: list[BuilderSource] = []

    if isinstance(
        function,
        (
            SingleShootingPrimalFunction,
            SingleShootingGradientFunction,
            SingleShootingHvpFunction,
            SingleShootingJointFunction,
        ),
    ):
        for request in requests:
            if request.kind != "primal":
                raise ValueError(
                    "explicit single-shooting staged sources in CodeGenerationBuilder "
                    "currently support only add_primal()"
                )
            if request.components:
                raise ValueError("include_states is configured on the single-shooting staged source itself")
            resolved.append(
                _rename_builder_source(
                    replace(function, simplification=simplification),
                    _builder_function_name(
                        crate_prefix,
                        "f",
                        base_name=base_name,
                        include_base_name=include_base_name,
                    ),
                )
            )
        return tuple(resolved)

    simplified_problem = replace(function, simplification=simplification)
    for request in requests:
        if request.kind == "primal":
            include_states = "states" in request.components
            labels = ("f", "states") if include_states else ("f",)
            resolved.append(
                _rename_builder_source(
                    simplified_problem.primal(include_states=include_states),
                    _builder_function_name(
                        crate_prefix,
                        *labels,
                        base_name=base_name,
                        include_base_name=include_base_name,
                    ),
                )
            )
            continue
        if request.kind == "gradient":
            include_states = "states" in request.components
            gradient_source = simplified_problem.gradient(include_states=include_states)
            labels = ("grad", "states") if include_states else ("grad",)
            resolved.append(
                _rename_builder_source(
                    gradient_source,
                    _builder_function_name(
                        crate_prefix,
                        *labels,
                        base_name=base_name,
                        include_base_name=include_base_name,
                        input_name=simplified_problem.control_sequence_name,
                        include_input_name=True,
                    ),
                )
            )
            continue
        if request.kind == "hvp":
            include_states = "states" in request.components
            hvp_source = simplified_problem.hvp(include_states=include_states)
            labels = ("hvp", "states") if include_states else ("hvp",)
            resolved.append(
                _rename_builder_source(
                    hvp_source,
                    _builder_function_name(
                        crate_prefix,
                        *labels,
                        base_name=base_name,
                        include_base_name=include_base_name,
                        input_name=simplified_problem.control_sequence_name,
                        include_input_name=True,
                    ),
                )
            )
            continue
        if request.kind == "joint":
            if request.bundle is None or not isinstance(request.bundle, SingleShootingBundle):
                raise ValueError(
                    "joint single-shooting requests require a SingleShootingBundle"
                )
            labels: list[str] = []
            if request.bundle.include_cost:
                labels.append("f")
            if request.bundle.include_gradient:
                labels.append("grad")
            if request.bundle.include_hvp:
                labels.append("hvp")
            if request.bundle.include_states:
                labels.append("states")
            resolved.append(
                _rename_builder_source(
                    simplified_problem.joint(request.bundle),
                    _builder_function_name(
                        crate_prefix,
                        *labels,
                        base_name=base_name,
                        include_base_name=include_base_name,
                        input_name=simplified_problem.control_sequence_name,
                        include_input_name=request.bundle.include_gradient or request.bundle.include_hvp,
                    ),
                )
            )
            continue
        raise ValueError(
            "single-shooting sources currently support only add_primal(), "
            "add_gradient(), add_hvp(), and add_joint()"
        )

    return tuple(resolved)


def _apply_builder_base_name(function: Function, function_name: str | None) -> Function:
    if function_name is None or function_name == function.name:
        return function
    return Function(
        function_name,
        function.inputs,
        function.outputs,
        input_names=function.input_names,
        output_names=function.output_names,
    )


def _rename_generated_function(function: Function, name: str) -> Function:
    if function.name == name:
        return function
    return Function(
        name,
        function.inputs,
        function.outputs,
        input_names=function.input_names,
        output_names=function.output_names,
    )


def _rename_builder_source(
    function: BuilderSource,
    name: str,
) -> BuilderSource:
    """Return a source object with an updated generated Rust function name."""
    if function.name == name:
        return function
    if isinstance(function, Function):
        return _rename_generated_function(function, name)
    return replace(function, name=name)


def _builder_joint_labels(components: tuple[str, ...]) -> tuple[str, ...]:
    mapping = {
        "f": "f",
        "grad": "grad",
        "jf": "jf",
        "hessian": "hessian",
        "hvp": "hvp",
    }
    return tuple(mapping[component] for component in components)


def _builder_function_name(
    crate_prefix: str,
    *labels: str,
    base_name: str | None = None,
    include_base_name: bool = False,
    input_name: str | None = None,
    include_input_name: bool = False,
) -> str:
    parts = [crate_prefix]
    if include_base_name and base_name is not None:
        parts.append(sanitize_ident(base_name))
    parts.extend(labels)
    if include_input_name and input_name is not None:
        parts.append(sanitize_ident(input_name))
    return "_".join(parts)


def _normalize_wrt_indices(wrt: int | list[int] | tuple[int, ...]) -> tuple[int, ...]:
    if isinstance(wrt, int):
        return (wrt,)
    return tuple(wrt)


def _resolve_function_bundle(
    bundle: FunctionBundle,
    input_count: int,
) -> tuple[tuple[int, tuple[str, ...]], ...]:
    if not bundle.items:
        raise ValueError("FunctionBundle must contain at least one item")

    has_primal = any(item.kind == "f" for item in bundle.items)
    grouped: dict[int, list[str]] = {}

    for item in bundle.items:
        if item.kind == "f":
            continue
        if item.wrt_indices is None:
            raise ValueError(f"FunctionBundle item {item.kind!r} requires a wrt specification")
        for index in item.wrt_indices:
            if not 0 <= index < input_count:
                raise IndexError("wrt_index is out of range")
            grouped.setdefault(index, [])
            if has_primal and "f" not in grouped[index]:
                grouped[index].append("f")
            if item.kind not in grouped[index]:
                grouped[index].append(item.kind)

    if not grouped:
        raise ValueError("FunctionBundle must include at least one derivative-producing item")

    resolved: list[tuple[int, tuple[str, ...]]] = []
    for index in sorted(grouped):
        components = tuple(grouped[index])
        if len(components) < 2:
            raise ValueError("joint functions require at least two components")
        resolved.append((index, components))
    return tuple(resolved)


def _maybe_simplify_generated_function(
    function: Function,
    simplification: int | str | None,
) -> Function:
    if simplification is None:
        return function
    return function.simplify(max_effort=simplification, name=function.name)
